"""
This module is for creating dplms dictionary for ge processing
"""

from __future__ import annotations

import itertools
import json
import logging
import os
import time

import matplotlib.pyplot as plt
import numpy as np
from lgdo import Array, Table, lh5
from scipy.signal import convolve, convolve2d

from pygama.math.histogram import get_hist
from pygama.math.peak_fitting import (
    extended_gauss_step_pdf,
    extended_radford_pdf,
    gauss_step_pdf,
    radford_pdf,
)
from pygama.pargen.cuts import generate_cuts, get_cut_indexes
from pygama.pargen.dsp_optimize import run_one_dsp
from pygama.pargen.energy_optimisation import fom_FWHM_with_dt_corr_fit

log = logging.getLogger(__name__)
sto = lh5.LH5Store()


def dplms_ge_dict(
    lh5_path: str,
    raw_fft: Table,
    raw_cal: Table,
    dsp_config: dict,
    par_dsp: dict,
    par_dsp_lh5: str,
    dplms_dict: dict,
    decay_const: float = 0,
    ene_par: str = "dplmsEmax",
    display: int = 0,
) -> dict:
    """
    This function calculates the dplms dictionary for HPGe detectors.

    Parameters
    ----------
    lh5_path
        Name of channel to process, should be name of lh5 group in raw files
    fft_files
        table with fft data
    raw_cal
        table with cal data
    dsp_config
        dsp config file
    par_dsp
        Dictionary with db parameters for dsp processing
    par_dsp_lh5
        Path for saving dplms coefficients
    dplms_dict
        Dictionary with various parameters

    Returns
    -------
    out_dict
    """

    t0 = time.time()
    log.info(f"\nSelecting baselines")

    dsp_fft = run_one_dsp(raw_fft, dsp_config, db_dict=par_dsp[lh5_path])
    cut_dict = generate_cuts(dsp_fft, parameters=dplms_dict["bls_cut_pars"])
    idxs = get_cut_indexes(dsp_fft, cut_dict)
    bl_field = dplms_dict["bl_field"]
    log.info(f"... {len(dsp_fft[bl_field].values.nda[idxs,:])} baselines after cuts")

    bls = dsp_fft[bl_field].values.nda[idxs, : dplms_dict["bsize"]]
    bls_par = {}
    bls_cut_pars = [par for par in dplms_dict["bls_cut_pars"].keys()]
    for par in bls_cut_pars:
        bls_par[par] = dsp_fft[par].nda
    t1 = time.time()
    log.info(
        f"total events {len(raw_fft)}, {len(bls)} baseline selected in {(t1-t0):.2f} s"
    )

    log.info(
        "\nCalculating noise matrix of length",
        dplms_dict["length"],
        "n. events",
        bls.shape[0],
        "size",
        bls.shape[1],
    )
    nmat = noise_matrix(bls, dplms_dict["length"])
    t2 = time.time()
    log.info(f"Time to calculate noise matrix {(t2-t1):.2f} s")

    log.info("\nSelecting signals")
    wsize = dplms_dict["wsize"]
    wf_field = dplms_dict["wf_field"]
    peaks_keV = np.array(dplms_dict["peaks_keV"])
    kev_widths = [tuple(kev_width) for kev_width in dplms_dict["kev_widths"]]

    log.info(f"Produce dsp data for {len(raw_cal)} events")
    dsp_cal = run_one_dsp(raw_cal, dsp_config, db_dict=par_dsp[lh5_path])
    t3 = time.time()
    log.info(f"Time to run dsp production {(t3-t2):.2f} s")

    dsp_config["outputs"] = [ene_par, "dt_eff"]

    # dictionary for peak fitting
    peak_dict = {
        "peak": peaks_keV[-1],
        "kev_width": kev_widths[-1],
        "parameter": ene_par,
        "func": extended_gauss_step_pdf,
        "gof_func": gauss_step_pdf,
    }

    if display > 0:
        plot_dict = {}
        plot_dict["dplms"] = {}

    # penalized coefficients
    dp_coeffs = dplms_dict["dp_coeffs"]
    za_coeff = dplms_dict["dp_def"]["za"]
    dp_coeffs.pop("za")
    coeff_keys = [key for key in dp_coeffs.keys()]
    lists = [dp_coeffs[key] for key in dp_coeffs.keys()]

    prod = list(itertools.product(*lists))
    grid_dict = {}
    min_fom = float("inf")
    min_idx = None

    for i, values in enumerate(prod):
        coeff_values = dict(zip(coeff_keys, values))

        log.info(
            "\nCase",
            i,
            "->",
            ", ".join(f"{key} = {value}" for key, value in coeff_values.items()),
        )
        grid_dict[i] = coeff_values

        sel_dict = signal_selection(dsp_cal, dplms_dict, coeff_values)
        wfs = dsp_cal[wf_field].nda[sel_dict["idxs"], :]
        log.info(f"... {len(wfs)} signals after signal selection")

        ref, rmat, pmat, fmat = signal_matrices(wfs, dplms_dict["length"], decay_const)

        t_tmp = time.time()
        nm_coeff = coeff_values["nm"]
        ft_coeff = coeff_values["ft"]
        x, y, refy = filter_synthesis(
            ref,
            nm_coeff * nmat,
            rmat,
            za_coeff,
            pmat,
            ft_coeff * fmat,
            dplms_dict["length"],
            wsize,
        )
        par_dsp[lh5_path]["dplms"] = {"length": dplms_dict["length"], "coefficients": x}
        log.info(
            f"Filter synthesis in {time.time()-t_tmp:.1f} s, filter area", np.sum(x)
        )

        t_tmp = time.time()
        dsp_opt = run_one_dsp(raw_cal, dsp_config, db_dict=par_dsp[lh5_path])

        try:
            res = fom_FWHM_with_dt_corr_fit(
                dsp_opt,
                peak_dict,
                "QDrift",
                idxs=np.where(~np.isnan(dsp_opt["dt_eff"].nda))[0],
            )
        except:
            log.debug("FWHM not calculated")
            continue

        fwhm, fwhm_err, alpha, chisquare = (
            res["fwhm"],
            res["fwhm_err"],
            res["alpha"],
            res["chisquare"],
        )
        log.info(
            f"FWHM = {fwhm:.2f} ± {fwhm_err:.2f} keV, evaluated in {time.time()-t_tmp:.1f} s"
        )

        grid_dict[i]["fwhm"] = fwhm
        grid_dict[i]["fwhm_err"] = fwhm_err
        grid_dict[i]["alpha"] = alpha

        if (
            fwhm < dplms_dict["fwhm_limit"]
            and fwhm_err < dplms_dict["err_limit"]
            and chisquare < dplms_dict["chi_limit"]
        ):
            if fwhm < min_fom:
                min_idx, min_fom = i, fwhm

    if min_idx is not None:
        min_result = grid_dict[min_idx]
        best_case_values = {key: min_result[key] for key in min_result.keys()}

        fwhm = best_case_values.get("fwhm", None)
        fwhm_err = best_case_values.get("fwhm_err", 0)
        alpha = best_case_values.get("alpha", 0)
        nm_coeff = best_case_values.get("nm", dplms_dict["dp_def"]["nm"])
        ft_coeff = best_case_values.get("ft", dplms_dict["dp_def"]["nm"])
        rt_coeff = best_case_values.get("rt", dplms_dict["dp_def"]["rt"])
        pt_coeff = best_case_values.get("pt", dplms_dict["dp_def"]["pt"])

        if all(
            v is not None
            for v in [
                fwhm,
                fwhm_err,
                alpha,
                nm_coeff,
                ft_coeff,
                rt_coeff,
                pt_coeff,
            ]
        ):
            log.info(
                f"\nBest case: FWHM = {fwhm:.2f} ± {fwhm_err:.2f} keV, ctc {alpha}"
            )
        else:
            log.error("Some values are missing in the best case results")
    else:
        log.error("Filter synthesis failed")
        nm_coeff = dplms_dict["dp_def"]["nm"]
        ft_coeff = dplms_dict["dp_def"]["ft"]
        rt_coeff = dplms_dict["dp_def"]["rt"]
        pt_coeff = dplms_dict["dp_def"]["pt"]

    # filter synthesis
    sel_dict = signal_selection(dsp_cal, dplms_dict, best_case_values)
    idxs = sel_dict["idxs"]
    wfs = dsp_cal[wf_field].nda[idxs, :]
    ref, rmat, pmat, fmat = signal_matrices(wfs, dplms_dict["length"], decay_const)

    x, y, refy = filter_synthesis(
        ref,
        nm_coeff * nmat,
        rmat,
        za_coeff,
        pmat,
        ft_coeff * fmat,
        dplms_dict["length"],
        wsize,
    )

    sto.write(
        Array(x),
        name="dplms",
        lh5_file=par_dsp_lh5,
        wo_mode="overwrite",
        group=lh5_path,
    )

    out_dict = {
        "dplms": {
            "length": dplms_dict["length"],
            "coefficients": f"loadlh5('{par_dsp_lh5}', '{lh5_path}/dplms')",
            "dp_coeffs": {
                "nm": nm_coeff,
                "za": za_coeff,
                "ft": ft_coeff,
                "rt": rt_coeff,
                "pt": pt_coeff,
            },
        }
    }
    out_alpha_dict = {
        f"{ene_par}_ctc": {
            "expression": f"{ene_par}*(1+dt_eff*a)",
            "parameters": {"a": round(alpha, 9)},
        }
    }
    out_dict.update({"ctc_params": out_alpha_dict})

    log.info(f"Time to complete DPLMS filter synthesis {time.time()-t0:.1f}")

    if display > 0:
        plot_dict["dplms"]["ref"] = ref
        plot_dict["dplms"]["coefficients"] = x

        bl_idxs = np.random.choice(len(bls), dplms_dict["n_plot"])
        bls = bls[bl_idxs]
        fig, ax = plt.subplots(figsize=(12, 6.75), facecolor="white")
        for ii, wf in enumerate(bls):
            if ii < 10:
                ax.plot(wf, label=f"mean = {wf.mean():.1f}")
            else:
                ax.plot(wf)
        ax.legend(title=f"{lh5_path}", loc="upper right")
        plot_dict["dplms"]["bls"] = fig
        fig, ax = plt.subplots(nrows=2, ncols=3, figsize=(16, 9), facecolor="white")
        for ii, par in enumerate(bls_cut_pars):
            mean = cut_dict[par]["Mean Value"]
            llo, lup = cut_dict[par]["Lower Boundary"], cut_dict[par]["Upper Boundary"]
            plo, pup = mean - 2 * (mean - llo), mean + 2 * (lup - mean)
            hh, bb = np.histogram(bls_par[par], bins=np.linspace(plo, pup, 200))
            ax.flat[ii].plot(bb[1:], hh, ds="steps", label=f"cut on {par}")
            ax.flat[ii].axvline(lup, color="k", linestyle=":", label="selection")
            ax.flat[ii].axvline(llo, color="k", linestyle=":")
            ax.flat[ii].set_xlabel(par)
            ax.flat[ii].set_yscale("log")
            ax.flat[ii].legend(title=f"{lh5_path}", loc="upper right")
        plot_dict["dplms"]["bl_sel"] = fig

        wf_idxs = np.random.choice(len(wfs), dplms_dict["n_plot"])
        wfs = wfs[wf_idxs]
        peak_pos = dsp_cal["peak_pos"].nda
        peak_pos_neg = dsp_cal["peak_pos_neg"].nda
        centroid = dsp_cal["centroid"].nda
        risetime = dsp_cal["tp_90"].nda - dsp_cal["tp_10"].nda
        rt_low = dplms_dict["rt_low"]
        rt_high = dplms_dict["rt_high"]
        peak_lim = dplms_dict["peak_lim"]
        cal_par = {}
        wfs_cut_pars = [par for par in dplms_dict["wfs_cut_pars"].keys()]
        for par in wfs_cut_pars:
            cal_par[par] = dsp_cal[par].nda
        fig, ax = plt.subplots(figsize=(12, 6.75), facecolor="white")
        for ii, wf in enumerate(wfs):
            if ii < 10:
                ax.plot(wf, label=f"centr = {centroid[ii]}")
            else:
                ax.plot(wf)
        ax.legend(title=f"{lh5_path}", loc="upper right")
        axin = ax.inset_axes([0.1, 0.15, 0.35, 0.5])
        for wf in wfs:
            axin.plot(wf)
        axin.set_xlim(wsize / 2 - dplms_dict["zoom"], wsize / 2 + dplms_dict["zoom"])
        axin.set_yticklabels("")
        plot_dict["dplms"]["wfs"] = fig
        fig, ax = plt.subplots(nrows=2, ncols=3, figsize=(16, 9), facecolor="white")
        wfs_cut_pars.append("centroid")
        wfs_cut_pars.append("peak_pos")
        wfs_cut_pars.append("risetime")
        for ii, par in enumerate(wfs_cut_pars):
            pspace = np.linspace(
                wsize / 2 - peak_lim, wsize / 2 + peak_lim, 2 * peak_lim
            )
            if par == "centroid":
                llo, lup = sel_dict["ct_ll"], sel_dict["ct_hh"]
                hh, bb = np.histogram(centroid, bins=pspace)
            elif par == "peak_pos":
                llo, lup = sel_dict["pp_ll"], sel_dict["pp_hh"]
                hh, bb = np.histogram(peak_pos, bins=pspace)
            elif par == "risetime":
                llo, lup = sel_dict["rt_ll"], sel_dict["rt_hh"]
                rt_bins = int((rt_high - rt_low) / dplms_dict["period"])
                rt_space = np.linspace(rt_low, rt_high, rt_bins)
                hh, bb = np.histogram(risetime, bins=rt_space)
            else:
                llo, lup = np.min(cal_par[par]), np.max(cal_par[par])
                hh, bb = np.histogram(cal_par[par], bins=np.linspace(llo, lup, 200))
            ax.flat[ii + 1].plot(bb[1:], hh, ds="steps", label=f"cut on {par}")
            ax.flat[ii + 1].axvline(
                llo, color="k", linestyle=":", label=f"sel. {llo:.1f} {lup:.1f}"
            )
            if par != "centroid":
                ax.flat[ii + 1].axvline(lup, color="k", linestyle=":")
            ax.flat[ii + 1].set_xlabel(par)
            ax.flat[ii + 1].set_yscale("log")
            ax.flat[ii + 1].legend(title=f"{lh5_path}", loc="upper right")
        roughenergy = dsp_cal["trapTmax"].nda
        roughenergy_sel = roughenergy[idxs]
        ell, ehh = roughenergy.min(), roughenergy.max()
        he, be = np.histogram(roughenergy, bins=np.linspace(ell, ehh, 1000))
        hs, be = np.histogram(roughenergy_sel, bins=np.linspace(ell, ehh, 1000))
        ax.flat[0].plot(be[1:], he, c="b", ds="steps", label="initial")
        ax.flat[0].plot(be[1:], hs, c="r", ds="steps", label="selected")
        ax.flat[0].set_xlabel("rough energy (ADC)")
        ax.flat[0].set_yscale("log")
        ax.flat[0].legend(loc="upper right", title=f"{lh5_path}")
        plot_dict["dplms"]["wf_sel"] = fig

        fig, ax = plt.subplots(figsize=(12, 6.75), facecolor="white")
        ax.plot(x, "r-", label=f"filter")
        ax.axhline(0, color="black", linestyle=":")
        ax.legend(loc="upper right", title=f"{lh5_path}")
        axin = ax.inset_axes([0.6, 0.1, 0.35, 0.33])
        axin.plot(x, "r-")
        axin.set_xlim(
            dplms_dict["length"] / 2 - dplms_dict["zoom"],
            dplms_dict["length"] / 2 + dplms_dict["zoom"],
        )
        axin.set_yticklabels("")
        ax.indicate_inset_zoom(axin)

        return out_dict, plot_dict
    else:
        return out_dict


def is_valid_centroid(
    centroid: np.array, lim: int, size: int, full_size: int
) -> list[bool]:
    llim = size / 2 - lim
    hlim = full_size - size / 2
    idxs = (centroid > llim) & (centroid < hlim)
    return idxs, llim, hlim


def is_not_pile_up(
    peak_pos: np.array, peak_pos_neg: np.array, thr: int, lim: int, size: int
) -> list[bool]:
    bin_edges = np.linspace(size / 2 - lim, size / 2 + lim, 2 * lim)
    hist, bin_edges = np.histogram(peak_pos, bins=bin_edges)

    thr = thr * hist.max() / 100
    low_thr_idxs = np.where(hist[: hist.argmax()] < thr)[0]
    upp_thr_idxs = np.where(hist[hist.argmax() :] < thr)[0]

    idx_low = low_thr_idxs[-1] if low_thr_idxs.size > 0 else 0
    idx_upp = (
        upp_thr_idxs[0] + hist.argmax() if upp_thr_idxs.size > 0 else len(hist) - 1
    )

    llow, lupp = bin_edges[idx_low], bin_edges[idx_upp]

    idxs = []
    for n, nn in zip(peak_pos, peak_pos_neg):
        condition1 = np.count_nonzero(n > 0) == 1
        condition2 = (
            np.count_nonzero((n > 0) & ((n < llow) | (n > lupp) & (n < size))) == 0
        )
        condition3 = np.count_nonzero(nn > 0) == 0
        idxs.append(condition1 and condition2 and condition3)
    return idxs, llow, lupp


def is_valid_risetime(risetime: np.array, llim: int, perc: float):
    hlim = np.percentile(risetime[~np.isnan(risetime)], perc)
    idxs = (risetime >= llim) & (risetime <= hlim)
    return idxs, llim, hlim


def signal_selection(dsp_cal, dplms_dict, coeff_values):
    peak_pos = dsp_cal["peak_pos"].nda
    peak_pos_neg = dsp_cal["peak_pos_neg"].nda
    centroid = dsp_cal["centroid"].nda
    risetime = dsp_cal["tp_90"].nda - dsp_cal["tp_10"].nda

    rt_low = dplms_dict["rt_low"]
    rt_high = dplms_dict["rt_high"]
    peak_lim = dplms_dict["peak_lim"]
    wsize = dplms_dict["wsize"]
    bsize = dplms_dict["bsize"]

    centroid_lim = dplms_dict["centroid_lim"]
    if "rt" in coeff_values:
        perc = coeff_values["rt"]
    else:
        perc = dplms_dict["dp_def"]["rt"]
    if "pt" in coeff_values:
        thr = coeff_values["pt"]
    else:
        thr = dplms_dict["dp_def"]["rt"]

    idxs_ct, ct_ll, ct_hh = is_valid_centroid(centroid, centroid_lim, wsize, bsize)
    log.info(f"... {len(peak_pos[idxs_ct,:])} signals after alignment")

    idxs_pp, pp_ll, pp_hh = is_not_pile_up(peak_pos, peak_pos_neg, thr, peak_lim, wsize)
    log.info(f"... {len(peak_pos[idxs_pp,:])} signals after pile-up cut")

    idxs_rt, rt_ll, rt_hh = is_valid_risetime(risetime, rt_low, perc)
    log.info(f"... {len(peak_pos[idxs_rt,:])} signals after risetime cut")

    idxs = idxs_ct & idxs_pp & idxs_rt
    sel_dict = {
        "idxs": idxs,
        "ct_ll": ct_ll,
        "ct_hh": ct_hh,
        "pp_ll": pp_ll,
        "pp_hh": pp_hh,
        "rt_ll": rt_ll,
        "rt_hh": rt_hh,
    }
    return sel_dict


def noise_matrix(bls: np.array, length: int) -> np.array:
    nev, size = bls.shape
    ref = np.mean(bls, axis=0)
    offset = np.mean(ref)
    bls = bls - offset
    nmat = np.matmul(bls.T, bls, dtype=float) / nev
    kernel = np.identity(size - length + 1)
    nmat = convolve2d(nmat, kernel, boundary="symm", mode="valid") / (size - length + 1)
    return nmat


def signal_matrices(
    wfs: np.array, length: int, decay_const: float, ff: int = 2
) -> np.array:
    nev, size = wfs.shape
    lo = size // 2 - 100
    flo = size // 2 - length // 2
    fhi = size // 2 + length // 2
    offsets = np.mean(wfs[:, :lo], axis=1)
    wfs = wfs - offsets[:, np.newaxis]

    # Reference signal
    ref = np.sum(wfs, axis=0)
    ref /= np.max(ref)
    rmat = np.outer(ref[flo:fhi], ref[flo:fhi])

    # Pile-up matrix
    if decay_const > 0:
        decay = np.exp(-np.arange(length) / decay_const)
    else:
        decay = np.zeros(length)
    pmat = np.outer(decay, decay)

    # Flat top matrix
    flo -= ff // 2
    fhi += ff // 2
    wfs = wfs[:, flo:fhi]
    fmat = np.matmul(wfs.T, wfs, dtype=float) / nev
    m1 = ((1, -1), (-1, 1))
    fmat = convolve2d(fmat, m1, boundary="symm", mode="valid")
    if ff > 0:
        fmat = convolve2d(fmat, np.identity(ff), boundary="symm", mode="valid") / ff
    return ref, rmat, pmat, fmat


def filter_synthesis(
    ref: np.array,
    nmat: np.array,
    rmat: np.array,
    za: int,
    pmat: np.array,
    fmat: np.array,
    length: int,
    size: int,
    flip: bool = True,
) -> np.array:
    mat = nmat + rmat + za * np.ones([length, length]) + pmat + fmat
    flo = (size // 2) - (length // 2)
    fhi = (size // 2) + (length // 2)
    x = np.linalg.solve(mat, ref[flo:fhi]).astype(np.float32)
    y = convolve(ref, np.flip(x), mode="valid")
    maxy = np.max(y)
    x /= maxy
    y /= maxy
    refy = ref[(size // 2) - (len(y) // 2) : (size // 2) + (len(y) // 2)]
    if flip:
        return np.flip(x), y, refy
    else:
        return x, y, refy



"""
This module provides aggregators to build the `evt` tier.
"""

from __future__ import annotations

import awkward as ak
import numpy as np
from lgdo import Array, ArrayOfEqualSizedArrays, VectorOfVectors, lh5
from lgdo.lh5 import LH5Store
from numpy.typing import NDArray

from . import utils


def evaluate_to_first_or_last(
    cumulength: NDArray,
    idx: NDArray,
    ids: NDArray,
    f_hit: str,
    f_dsp: str,
    chns: list,
    chns_rm: list,
    expr: str,
    exprl: list,
    qry: str | NDArray,
    nrows: int,
    sorter: tuple,
    var_ph: dict = None,
    defv: bool | int | float = np.nan,
    is_first: bool = True,
    tcm_id_table_pattern: str = "ch{}",
    evt_group: str = "evt",
    hit_group: str = "hit",
    dsp_group: str = "dsp",
) -> Array:
    """Aggregates across channels by returning the expression of the channel
    with value of `sorter`.

    Parameters
    ----------
    idx
       `tcm` index array.
    ids
       `tcm` id array.
    f_hit
       path to `hit` tier file.
    f_dsp
       path to `dsp` tier file.
    chns
       list of channels to be aggregated.
    chns_rm
       list of channels to be skipped from evaluation and set to default value.
    expr
       expression string to be evaluated.
    exprl
       list of `dsp/hit/evt` parameter tuples in expression ``(tier, field)``.
    qry
       query expression to mask aggregation.
    nrows
       length of output array.
    sorter
       tuple of field in `hit/dsp/evt` tier to evaluate ``(tier, field)``.
    var_ph
       dictionary of `evt` and additional parameters and their values.
    defv
       default value.
    is_first
       defines if sorted by smallest or largest value of `sorter`
    tcm_id_table_pattern
        pattern to format `tcm` id values to table name in higher tiers. Must have one
        placeholder which is the `tcm` id.
    dsp_group
        LH5 root group in `dsp` file.
    hit_group
        LH5 root group in `hit` file.
    evt_group
        LH5 root group in `evt` file.
    """

    # define dimension of output array
    out = np.full(nrows, defv, dtype=type(defv))
    outt = np.zeros(len(out))

    store = LH5Store()

    for ch in chns:
        # get index list for this channel to be loaded
        idx_ch = idx[ids == utils.get_tcm_id_by_pattern(tcm_id_table_pattern, ch)]
        evt_ids_ch = np.searchsorted(
            cumulength,
            np.where(ids == utils.get_tcm_id_by_pattern(tcm_id_table_pattern, ch))[0],
            "right",
        )

        # evaluate at channel
        res = utils.get_data_at_channel(
            ch=ch,
            ids=ids,
            idx=idx,
            expr=expr,
            exprl=exprl,
            var_ph=var_ph,
            is_evaluated=ch not in chns_rm,
            f_hit=f_hit,
            f_dsp=f_dsp,
            defv=defv,
            tcm_id_table_pattern=tcm_id_table_pattern,
            evt_group=evt_group,
            hit_group=hit_group,
            dsp_group=dsp_group,
        )

        # get mask from query
        limarr = utils.get_mask_from_query(
            qry=qry,
            length=len(res),
            ch=ch,
            idx_ch=idx_ch,
            f_hit=f_hit,
            f_dsp=f_dsp,
            hit_group=hit_group,
            dsp_group=dsp_group,
        )

        # find if sorter is in hit or dsp
        t0 = store.read(
            f"{ch}/{sorter[0]}/{sorter[1]}",
            f_hit if f"{hit_group}" == sorter[0] else f_dsp,
            idx=idx_ch,
        )[0].view_as("np")

        if t0.ndim > 1:
            raise ValueError(f"sorter '{sorter[0]}/{sorter[1]}' must be a 1D array")

        if is_first:
            if ch == chns[0]:
                outt[:] = np.inf

            out[evt_ids_ch] = np.where(
                (t0 < outt[evt_ids_ch]) & (limarr), res, out[evt_ids_ch]
            )
            outt[evt_ids_ch] = np.where(
                (t0 < outt[evt_ids_ch]) & (limarr), t0, outt[evt_ids_ch]
            )

        else:
            out[evt_ids_ch] = np.where(
                (t0 > outt[evt_ids_ch]) & (limarr), res, out[evt_ids_ch]
            )
            outt[evt_ids_ch] = np.where(
                (t0 > outt[evt_ids_ch]) & (limarr), t0, outt[evt_ids_ch]
            )

    return Array(nda=out, dtype=type(defv))


def evaluate_to_scalar(
    mode: str,
    cumulength: NDArray,
    idx: NDArray,
    ids: NDArray,
    f_hit: str,
    f_dsp: str,
    chns: list,
    chns_rm: list,
    expr: str,
    exprl: list,
    qry: str | NDArray,
    nrows: int,
    var_ph: dict = None,
    defv: bool | int | float = np.nan,
    tcm_id_table_pattern: str = "ch{}",
    evt_group: str = "evt",
    hit_group: str = "hit",
    dsp_group: str = "dsp",
) -> Array:
    """Aggregates by summation across channels.

    Parameters
    ----------
    mode
       aggregation mode.
    idx
       `tcm` index array.
    ids
       `tcm` id array.
    f_hit
       path to `hit` tier file.
    f_dsp
       path to `dsp` tier file.
    chns
       list of channels to be aggregated.
    chns_rm
       list of channels to be skipped from evaluation and set to default value.
    expr
       expression string to be evaluated.
    exprl
       list of `dsp/hit/evt` parameter tuples in expression ``(tier, field)``.
    qry
       query expression to mask aggregation.
    nrows
       length of output array
    var_ph
       dictionary of `evt` and additional parameters and their values.
    defv
       default value.
    tcm_id_table_pattern
        pattern to format `tcm` id values to table name in higher tiers. Must have one
        placeholder which is the `tcm` id.
    dsp_group
        LH5 root group in `dsp` file.
    hit_group
        LH5 root group in `hit` file.
    evt_group
        LH5 root group in `evt` file.
    """

    # define dimension of output array
    out = np.full(nrows, defv, dtype=type(defv))

    for ch in chns:
        # get index list for this channel to be loaded
        idx_ch = idx[ids == utils.get_tcm_id_by_pattern(tcm_id_table_pattern, ch)]
        evt_ids_ch = np.searchsorted(
            cumulength,
            np.where(ids == utils.get_tcm_id_by_pattern(tcm_id_table_pattern, ch))[0],
            "right",
        )

        res = utils.get_data_at_channel(
            ch=ch,
            ids=ids,
            idx=idx,
            expr=expr,
            exprl=exprl,
            var_ph=var_ph,
            is_evaluated=ch not in chns_rm,
            f_hit=f_hit,
            f_dsp=f_dsp,
            defv=defv,
            tcm_id_table_pattern=tcm_id_table_pattern,
            evt_group=evt_group,
            hit_group=hit_group,
            dsp_group=dsp_group,
        )

        # get mask from query
        limarr = utils.get_mask_from_query(
            qry=qry,
            length=len(res),
            ch=ch,
            idx_ch=idx_ch,
            f_hit=f_hit,
            f_dsp=f_dsp,
            hit_group=hit_group,
            dsp_group=dsp_group,
        )

        # switch through modes
        if "sum" == mode:
            if res.dtype == bool:
                res = res.astype(int)
            out[evt_ids_ch] = np.where(limarr, res + out[evt_ids_ch], out[evt_ids_ch])
        if "any" == mode:
            if res.dtype != bool:
                res = res.astype(bool)
            out[evt_ids_ch] = out[evt_ids_ch] | (res & limarr)
        if "all" == mode:
            if res.dtype != bool:
                res = res.astype(bool)
            out[evt_ids_ch] = out[evt_ids_ch] & res & limarr

    return Array(nda=out, dtype=type(defv))


def evaluate_at_channel(
    cumulength: NDArray,
    idx: NDArray,
    ids: NDArray,
    f_hit: str,
    f_dsp: str,
    chns_rm: list,
    expr: str,
    exprl: list,
    ch_comp: Array,
    var_ph: dict = None,
    defv: bool | int | float = np.nan,
    tcm_id_table_pattern: str = "ch{}",
    evt_group: str = "evt",
    hit_group: str = "hit",
    dsp_group: str = "dsp",
) -> Array:
    """Aggregates by evaluating the expression at a given channel.

    Parameters
    ----------
    idx
       `tcm` index array.
    ids
       `tcm` id array.
    f_hit
       path to `hit` tier file.
    f_dsp
       path to `dsp` tier file.
    chns_rm
       list of channels to be skipped from evaluation and set to default value.
    expr
       expression string to be evaluated.
    exprl
       list of `dsp/hit/evt` parameter tuples in expression ``(tier, field)``.
    ch_comp
       array of rawids at which the expression is evaluated.
    var_ph
       dictionary of `evt` and additional parameters and their values.
    defv
       default value.
    tcm_id_table_pattern
        pattern to format `tcm` id values to table name in higher tiers. Must have one
        placeholder which is the `tcm` id.
    dsp_group
        LH5 root group in `dsp` file.
    hit_group
        LH5 root group in `hit` file.
    evt_group
        LH5 root group in `evt` file.
    """

    out = np.full(len(ch_comp.nda), defv, dtype=type(defv))

    for ch in np.unique(ch_comp.nda.astype(int)):
        # skip default value
        if utils.get_table_name_by_pattern(tcm_id_table_pattern, ch) not in lh5.ls(
            f_hit
        ):
            continue
        idx_ch = idx[ids == ch]
        evt_ids_ch = np.searchsorted(cumulength, np.where(ids == ch)[0], "right")
        res = utils.get_data_at_channel(
            ch=utils.get_table_name_by_pattern(tcm_id_table_pattern, ch),
            ids=ids,
            idx=idx,
            expr=expr,
            exprl=exprl,
            var_ph=var_ph,
            is_evaluated=utils.get_table_name_by_pattern(tcm_id_table_pattern, ch)
            not in chns_rm,
            f_hit=f_hit,
            f_dsp=f_dsp,
            defv=defv,
            tcm_id_table_pattern=tcm_id_table_pattern,
            evt_group=evt_group,
            hit_group=hit_group,
            dsp_group=dsp_group,
        )

        out[evt_ids_ch] = np.where(ch == ch_comp.nda[idx_ch], res, out[evt_ids_ch])

    return Array(nda=out, dtype=type(defv))


def evaluate_at_channel_vov(
    cumulength: NDArray,
    idx: NDArray,
    ids: NDArray,
    f_hit: str,
    f_dsp: str,
    expr: str,
    exprl: list,
    ch_comp: VectorOfVectors,
    chns_rm: list,
    var_ph: dict = None,
    defv: bool | int | float = np.nan,
    tcm_id_table_pattern: str = "ch{}",
    evt_group: str = "evt",
    hit_group: str = "hit",
    dsp_group: str = "dsp",
) -> VectorOfVectors:
    """Same as :func:`evaluate_at_channel` but evaluates expression at non
    flat channels :class:`.VectorOfVectors`.

    Parameters
    ----------
    idx
       `tcm` index array.
    ids
       `tcm` id array.
    f_hit
       path to `hit` tier file.
    f_dsp
       path to `dsp` tier file.
    expr
       expression string to be evaluated.
    exprl
       list of `dsp/hit/evt` parameter tuples in expression ``(tier, field)``.
    ch_comp
       array of "rawid"s at which the expression is evaluated.
    chns_rm
       list of channels to be skipped from evaluation and set to default value.
    var_ph
       dictionary of `evt` and additional parameters and their values.
    defv
       default value.
    tcm_id_table_pattern
        pattern to format `tcm` id values to table name in higher tiers. Must have one
        placeholder which is the `tcm` id.
     dsp_group
        LH5 root group in `dsp` file.
    hit_group
        LH5 root group in `hit` file.
    evt_group
        LH5 root group in `evt` file.
    """

    # blow up vov to aoesa
    out = ak.Array([[] for _ in range(len(ch_comp))])

    chns = np.unique(ch_comp.flattened_data.nda).astype(int)
    ch_comp = ch_comp.view_as("ak")

    type_name = None
    for ch in chns:
        evt_ids_ch = np.searchsorted(cumulength, np.where(ids == ch)[0], "right")
        res = utils.get_data_at_channel(
            ch=utils.get_table_name_by_pattern(tcm_id_table_pattern, ch),
            ids=ids,
            idx=idx,
            expr=expr,
            exprl=exprl,
            var_ph=var_ph,
            is_evaluated=utils.get_table_name_by_pattern(tcm_id_table_pattern, ch)
            not in chns_rm,
            f_hit=f_hit,
            f_dsp=f_dsp,
            defv=defv,
            tcm_id_table_pattern=tcm_id_table_pattern,
            evt_group=evt_group,
            hit_group=hit_group,
            dsp_group=dsp_group,
        )

        # see in which events the current channel is present
        mask = ak.to_numpy(ak.any(ch_comp == ch, axis=-1), allow_missing=False)
        cv = np.full(len(ch_comp), np.nan)
        cv[evt_ids_ch] = res
        cv[~mask] = np.nan
        cv = ak.drop_none(ak.nan_to_none(ak.Array(cv)[:, None]))

        out = ak.concatenate((out, cv), axis=-1)

        if ch == chns[0]:
            type_name = res.dtype

    return VectorOfVectors(ak.values_astype(out, type_name), dtype=type_name)


def evaluate_to_aoesa(
    cumulength: NDArray,
    idx: NDArray,
    ids: NDArray,
    f_hit: str,
    f_dsp: str,
    chns: list,
    chns_rm: list,
    expr: str,
    exprl: list,
    qry: str | NDArray,
    nrows: int,
    var_ph: dict = None,
    defv: bool | int | float = np.nan,
    missv=np.nan,
    tcm_id_table_pattern: str = "ch{}",
    evt_group: str = "evt",
    hit_group: str = "hit",
    dsp_group: str = "dsp",
) -> ArrayOfEqualSizedArrays:
    """Aggregates by returning an :class:`.ArrayOfEqualSizedArrays` of evaluated
    expressions of channels that fulfill a query expression.

    Parameters
    ----------
    idx
       `tcm` index array.
    ids
       `tcm` id array.
    f_hit
       path to `hit` tier file.
    f_dsp
       path to `dsp` tier file.
    chns
       list of channels to be aggregated.
    chns_rm
       list of channels to be skipped from evaluation and set to default value.
    expr
       expression string to be evaluated.
    exprl
       list of `dsp/hit/evt` parameter tuples in expression ``(tier, field)``.
    qry
       query expression to mask aggregation.
    nrows
       length of output :class:`.VectorOfVectors`.
    ch_comp
       array of "rawid"s at which the expression is evaluated.
    var_ph
       dictionary of `evt` and additional parameters and their values.
    defv
       default value.
    missv
       missing value.
    sorter
       sorts the entries in the vector according to sorter expression.
    tcm_id_table_pattern
        pattern to format `tcm` id values to table name in higher tiers. Must have one
        placeholder which is the `tcm` id.
    dsp_group
        LH5 root group in `dsp` file.
    hit_group
        LH5 root group in `hit` file.
    evt_group
        LH5 root group in `evt` file.
    """
    # define dimension of output array
    out = np.full((nrows, len(chns)), missv)

    i = 0
    for ch in chns:
        idx_ch = idx[ids == utils.get_tcm_id_by_pattern(tcm_id_table_pattern, ch)]
        evt_ids_ch = np.searchsorted(
            cumulength,
            np.where(ids == utils.get_tcm_id_by_pattern(tcm_id_table_pattern, ch))[0],
            "right",
        )
        res = utils.get_data_at_channel(
            ch=ch,
            ids=ids,
            idx=idx,
            expr=expr,
            exprl=exprl,
            var_ph=var_ph,
            is_evaluated=ch not in chns_rm,
            f_hit=f_hit,
            f_dsp=f_dsp,
            defv=defv,
            tcm_id_table_pattern=tcm_id_table_pattern,
            evt_group=evt_group,
            hit_group=hit_group,
            dsp_group=dsp_group,
        )

        # get mask from query
        limarr = utils.get_mask_from_query(
            qry=qry,
            length=len(res),
            ch=ch,
            idx_ch=idx_ch,
            f_hit=f_hit,
            f_dsp=f_dsp,
            hit_group=hit_group,
            dsp_group=dsp_group,
        )

        out[evt_ids_ch, i] = np.where(limarr, res, out[evt_ids_ch, i])

        i += 1

    return ArrayOfEqualSizedArrays(nda=out)


def evaluate_to_vector(
    cumulength: NDArray,
    idx: NDArray,
    ids: NDArray,
    f_hit: str,
    f_dsp: str,
    chns: list,
    chns_rm: list,
    expr: str,
    exprl: list,
    qry: str | NDArray,
    nrows: int,
    var_ph: dict = None,
    defv: bool | int | float = np.nan,
    sorter: str = None,
    tcm_id_table_pattern: str = "ch{}",
    evt_group: str = "evt",
    hit_group: str = "hit",
    dsp_group: str = "dsp",
) -> VectorOfVectors:
    """Aggregates by returning a :class:`.VectorOfVector` of evaluated
    expressions of channels that fulfill a query expression.

    Parameters
    ----------
    idx
       `tcm` index array.
    ids
       `tcm` id array.
    f_hit
       path to `hit` tier file.
    f_dsp
       path to `dsp` tier file.
    chns
       list of channels to be aggregated.
    chns_rm
       list of channels to be skipped from evaluation and set to default value.
    expr
       expression string to be evaluated.
    exprl
       list of `dsp/hit/evt` parameter tuples in expression ``(tier, field)``.
    qry
       query expression to mask aggregation.
    nrows
       length of output :class:`.VectorOfVectors`.
    ch_comp
       array of "rawids" at which the expression is evaluated.
    var_ph
       dictionary of `evt` and additional parameters and their values.
    defv
       default value.
    sorter
       sorts the entries in the vector according to sorter expression.
       ``ascend_by:<hit|dsp.field>`` results in an vector ordered ascending,
       ``decend_by:<hit|dsp.field>`` sorts descending.
    tcm_id_table_pattern
        pattern to format `tcm` id values to table name in higher tiers. Must have one
        placeholder which is the `tcm` id.
     dsp_group
        LH5 root group in `dsp` file.
    hit_group
        LH5 root group in `hit` file.
    evt_group
        LH5 root group in `evt` file.
    """
    out = evaluate_to_aoesa(
        cumulength=cumulength,
        idx=idx,
        ids=ids,
        f_hit=f_hit,
        f_dsp=f_dsp,
        chns=chns,
        chns_rm=chns_rm,
        expr=expr,
        exprl=exprl,
        qry=qry,
        nrows=nrows,
        var_ph=var_ph,
        defv=defv,
        missv=np.nan,
        tcm_id_table_pattern=tcm_id_table_pattern,
        evt_group=evt_group,
        hit_group=hit_group,
        dsp_group=dsp_group,
    ).view_as("np")

    # if a sorter is given sort accordingly
    if sorter is not None:
        md, fld = sorter.split(":")
        s_val = evaluate_to_aoesa(
            cumulength=cumulength,
            idx=idx,
            ids=ids,
            f_hit=f_hit,
            f_dsp=f_dsp,
            chns=chns,
            chns_rm=chns_rm,
            expr=fld,
            exprl=[tuple(fld.split("."))],
            qry=None,
            nrows=nrows,
            missv=np.nan,
            tcm_id_table_pattern=tcm_id_table_pattern,
            evt_group=evt_group,
            hit_group=hit_group,
            dsp_group=dsp_group,
        ).view_as("np")
        if "ascend_by" == md:
            out = out[np.arange(len(out))[:, None], np.argsort(s_val)]

        elif "descend_by" == md:
            out = out[np.arange(len(out))[:, None], np.argsort(-s_val)]
        else:
            raise ValueError(
                "sorter values can only have 'ascend_by' or 'descend_by' prefixes"
            )

    return VectorOfVectors(
        ak.values_astype(ak.drop_none(ak.nan_to_none(ak.Array(out))), type(defv)),
        dtype=type(defv),
    )



"""
Module for special event level routines for SiPMs

functions must take as the first 8 args in order:
- path to the hit file
- path to the dsp <file
- path to the tcm file
- hit LH5 root group
- dsp LH5 root group
- tcm LH5 root group
- pattern to cast table names to tcm channel ids
- list of channels processed
additional parameters are free to the user and need to be defined in the JSON
"""


import awkward as ak
import numpy as np
from lgdo import Array, VectorOfVectors
from lgdo.lh5 import LH5Store

from pygama.evt import utils


# get an 1D akward array from 0 to 2D array
# casted by minimum of a 2D array
def cast_trigger(
    trgr,
    tdefault: float,
    length: int = None,
) -> ak.Array:
    if isinstance(trgr, Array):
        return ak.fill_none(ak.nan_to_none(trgr.view_as("ak")), tdefault)

    elif isinstance(trgr, (VectorOfVectors)):
        return ak.fill_none(
            ak.min(ak.fill_none(trgr.view_as("ak"), tdefault), axis=-1), tdefault
        )

    elif isinstance(trgr, ak.Array):
        if trgr.ndim == 1:
            return ak.fill_none(trgr, tdefault)
        elif trgr.ndim == 2:
            return ak.fill_none(ak.min(ak.fill_none(trgr, tdefault), axis=-1), tdefault)
        else:
            raise ValueError(f"Too many dimensions: {trgr.ndim}")
    elif isinstance(trgr, (float, int)) and isinstance(length, int):
        return ak.Array([trgr] * length)
    else:
        raise ValueError(f"Can't deal with t0 of type {type(trgr)}")


# get SiPM coincidence window mask
def get_spm_mask(
    lim: float, trgr: ak.Array, tmin: float, tmax: float, pe: ak.Array, times: ak.Array
) -> ak.Array:
    if trgr.ndim != 1:
        raise ValueError("trigger array muse be 1 dimensional!")
    if (len(trgr) != len(pe)) or (len(trgr) != len(times)):
        raise ValueError(
            f"All arrays must have same dimension across first axis len(pe)={len(pe)}, len(times)={len(times)}, len(trgr)={len(trgr)}"
        )

    tmi = trgr - tmin
    tma = trgr + tmax

    mask = (
        ((times * 16.0) < tma[:, None]) & ((times * 16.0) > tmi[:, None]) & (pe > lim)
    )
    return mask


# get LAr indices according to mask per event over all channels
# mode 0 -> return pulse indices
# mode 1 -> return tcm indices
# mode 2 -> return rawids
# mode 3 -> return tcm_idx
def get_masked_tcm_idx(
    f_hit,
    f_dsp,
    f_tcm,
    hit_group,
    dsp_group,
    tcm_group,
    tcm_id_table_pattern,
    chs,
    lim,
    trgr,
    tdefault,
    tmin,
    tmax,
    mode=0,
) -> VectorOfVectors:
    # load TCM data to define an event
    store = LH5Store()
    ids = store.read(f"/{tcm_group}/array_id", f_tcm)[0].view_as("np")
    idx = store.read(f"/{tcm_group}/array_idx", f_tcm)[0].view_as("np")

    arr_lst = []

    if isinstance(trgr, (float, int)):
        tge = cast_trigger(trgr, tdefault, length=np.max(idx) + 1)
    else:
        tge = cast_trigger(trgr, tdefault, length=None)

    for ch in chs:
        idx_ch = idx[ids == utils.get_tcm_id_by_pattern(tcm_id_table_pattern, ch)]

        pe = store.read(f"{ch}/{hit_group}/energy_in_pe", f_hit, idx=idx_ch)[0].view_as(
            "np"
        )
        tmp = np.full((np.max(idx) + 1, len(pe[0])), np.nan)
        tmp[idx_ch] = pe
        pe = ak.drop_none(ak.nan_to_none(ak.Array(tmp)))

        # times are in sample units
        times = store.read(f"{ch}/{hit_group}/trigger_pos", f_hit, idx=idx_ch)[
            0
        ].view_as("np")
        tmp = np.full((np.max(idx) + 1, len(times[0])), np.nan)
        tmp[idx_ch] = times
        times = ak.drop_none(ak.nan_to_none(ak.Array(tmp)))

        mask = get_spm_mask(lim, tge, tmin, tmax, pe, times)

        if mode == 0:
            out_idx = ak.local_index(mask)[mask]

        elif mode == 1:
            out_idx = np.full((np.max(idx) + 1), np.nan)
            out_idx[idx_ch] = np.where(
                ids == utils.get_tcm_id_by_pattern(tcm_id_table_pattern, ch)
            )[0]
            out_idx = ak.drop_none(ak.nan_to_none(ak.Array(out_idx)[:, None]))
            out_idx = out_idx[mask[mask] - 1]

        elif mode == 2:
            out_idx = ak.Array(
                [utils.get_tcm_id_by_pattern(tcm_id_table_pattern, ch)] * len(mask)
            )
            out_idx = out_idx[:, None][mask[mask] - 1]

        elif mode == 3:
            out_idx = np.full((np.max(idx) + 1), np.nan)
            out_idx[idx_ch] = idx_ch
            out_idx = ak.drop_none(ak.nan_to_none(ak.Array(out_idx)[:, None]))
            out_idx = out_idx[mask[mask] - 1]

        else:
            raise ValueError("Unknown mode")

        arr_lst.append(out_idx)

    return VectorOfVectors(array=ak.concatenate(arr_lst, axis=-1))


def get_spm_ene_or_maj(
    f_hit,
    f_tcm,
    hit_group,
    tcm_group,
    tcm_id_table_pattern,
    chs,
    lim,
    trgr,
    tdefault,
    tmin,
    tmax,
    mode,
):
    if mode not in ["energy_hc", "energy_dplms", "majority_hc", "majority_dplms"]:
        raise ValueError("Unknown mode")

    # load TCM data to define an event
    store = LH5Store()
    ids = store.read(f"/{tcm_group}/array_id", f_tcm)[0].view_as("np")
    idx = store.read(f"/{tcm_group}/array_idx", f_tcm)[0].view_as("np")
    out = np.zeros(np.max(idx) + 1)

    if isinstance(trgr, (float, int)):
        tge = cast_trigger(trgr, tdefault, length=np.max(idx) + 1)
    else:
        tge = cast_trigger(trgr, tdefault, length=None)

    for ch in chs:
        idx_ch = idx[ids == utils.get_tcm_id_by_pattern(tcm_id_table_pattern, ch)]

        if mode in ["energy_dplms", "majority_dplms"]:
            pe = ak.drop_none(
                ak.nan_to_none(
                    store.read(
                        f"{ch}/{hit_group}/energy_in_pe_dplms", f_hit, idx=idx_ch
                    )[0].view_as("ak")
                )
            )

            # times are in sample units
            times = ak.drop_none(
                ak.nan_to_none(
                    store.read(
                        f"{ch}/{hit_group}/trigger_pos_dplms", f_hit, idx=idx_ch
                    )[0].view_as("ak")
                )
            )

        else:
            pe = ak.drop_none(
                ak.nan_to_none(
                    store.read(f"{ch}/{hit_group}/energy_in_pe", f_hit, idx=idx_ch)[
                        0
                    ].view_as("ak")
                )
            )

            # times are in sample units
            times = ak.drop_none(
                ak.nan_to_none(
                    store.read(f"{ch}/{hit_group}/trigger_pos", f_hit, idx=idx_ch)[
                        0
                    ].view_as("ak")
                )
            )

        mask = get_spm_mask(lim, tge[idx_ch], tmin, tmax, pe, times)
        pe = pe[mask]

        if mode in ["energy_hc", "energy_dplms"]:
            out[idx_ch] = out[idx_ch] + ak.to_numpy(ak.nansum(pe, axis=-1))

        else:
            out[idx_ch] = out[idx_ch] + ak.to_numpy(
                ak.where(ak.nansum(pe, axis=-1) > lim, 1, 0)
            )

    return Array(nda=out)


# get LAr energy per event over all channels
def get_energy(
    f_hit,
    f_dsp,
    f_tcm,
    hit_group,
    dsp_group,
    tcm_group,
    tcm_id_table_pattern,
    chs,
    lim,
    trgr,
    tdefault,
    tmin,
    tmax,
) -> Array:
    return get_spm_ene_or_maj(
        f_hit,
        f_tcm,
        hit_group,
        tcm_group,
        tcm_id_table_pattern,
        chs,
        lim,
        trgr,
        tdefault,
        tmin,
        tmax,
        "energy_hc",
    )


# get LAr majority per event over all channels
def get_majority(
    f_hit,
    f_dsp,
    f_tcm,
    hit_group,
    dsp_group,
    tcm_group,
    tcm_id_table_pattern,
    chs,
    lim,
    trgr,
    tdefault,
    tmin,
    tmax,
) -> Array:
    return get_spm_ene_or_maj(
        f_hit,
        f_tcm,
        hit_group,
        tcm_group,
        tcm_id_table_pattern,
        chs,
        lim,
        trgr,
        tdefault,
        tmin,
        tmax,
        "majority_hc",
    )


# get LAr energy per event over all channels
def get_energy_dplms(
    f_hit,
    f_dsp,
    f_tcm,
    hit_group,
    dsp_group,
    tcm_group,
    tcm_id_table_pattern,
    chs,
    lim,
    trgr,
    tdefault,
    tmin,
    tmax,
) -> Array:
    return get_spm_ene_or_maj(
        f_hit,
        f_tcm,
        hit_group,
        tcm_group,
        tcm_id_table_pattern,
        chs,
        lim,
        trgr,
        tdefault,
        tmin,
        tmax,
        "energy_dplms",
    )


# get LAr majority per event over all channels
def get_majority_dplms(
    f_hit,
    f_dsp,
    f_tcm,
    hit_group,
    dsp_group,
    tcm_group,
    tcm_id_table_pattern,
    chs,
    lim,
    trgr,
    tdefault,
    tmin,
    tmax,
) -> Array:
    return get_spm_ene_or_maj(
        f_hit,
        f_tcm,
        hit_group,
        tcm_group,
        tcm_id_table_pattern,
        chs,
        lim,
        trgr,
        tdefault,
        tmin,
        tmax,
        "majority_dplms",
    )


# Calculate the ETC in different trailing modes:
# trail = 0: Singlet window = [tge,tge+swin]
# trail = 1: Singlet window = [t_first_lar_pulse, t_first_lar_pulse+ swin]
# trail = 2: Like trail = 1, but t_first_lar_pulse <= tge is ensured
# min_first_pls_ene sets the minimum energy of the first pulse (only used in trail > 0)
# max_per_channel, maximum number of pes a channel is allowed to have, if above it gets excluded
def get_etc(
    f_hit,
    f_dsp,
    f_tcm,
    hit_group,
    dsp_group,
    tcm_group,
    tcm_id_table_pattern,
    chs,
    lim,
    trgr,
    tdefault,
    tmin,
    tmax,
    swin,
    trail,
    min_first_pls_ene,
    max_per_channel,
) -> Array:
    # load TCM data to define an event
    store = LH5Store()
    ids = store.read(f"/{tcm_group}/array_id", f_tcm)[0].view_as("np")
    idx = store.read(f"/{tcm_group}/array_idx", f_tcm)[0].view_as("np")
    pe_lst = []
    time_lst = []

    if isinstance(trgr, (float, int)):
        tge = cast_trigger(trgr, tdefault, length=np.max(idx) + 1)
    else:
        tge = cast_trigger(trgr, tdefault, length=None)

    for ch in chs:
        idx_ch = idx[ids == utils.get_tcm_id_by_pattern(tcm_id_table_pattern, ch)]

        pe = store.read(f"{ch}/{hit_group}/energy_in_pe", f_hit, idx=idx_ch)[0].view_as(
            "np"
        )
        tmp = np.full((np.max(idx) + 1, len(pe[0])), np.nan)
        tmp[idx_ch] = pe
        pe = ak.drop_none(ak.nan_to_none(ak.Array(tmp)))

        # times are in sample units
        times = store.read(f"{ch}/{hit_group}/trigger_pos", f_hit, idx=idx_ch)[
            0
        ].view_as("np")
        tmp = np.full((np.max(idx) + 1, len(times[0])), np.nan)
        tmp[idx_ch] = times
        times = ak.drop_none(ak.nan_to_none(ak.Array(tmp)))

        mask = get_spm_mask(lim, tge, tmin, tmax, pe, times)

        pe = pe[mask]

        # max pe mask
        max_pe_mask = ak.nansum(pe, axis=-1) < max_per_channel
        pe = ak.drop_none(
            ak.nan_to_none(ak.where(max_pe_mask, pe, ak.Array([[np.nan]])))
        )
        pe_lst.append(pe)

        times = times[mask] * 16
        times = ak.drop_none(
            ak.nan_to_none(ak.where(max_pe_mask, times, ak.Array([[np.nan]])))
        )
        time_lst.append(times)

    pe_all = ak.concatenate(pe_lst, axis=-1)
    time_all = ak.concatenate(time_lst, axis=-1)

    if trail > 0:
        t1d = ak.min(time_all[pe_all > min_first_pls_ene], axis=-1)

        if trail == 2:
            t1d = ak.where(t1d > tge, tge, t1d)

        mask_total = time_all > t1d
        mask_singlet = (time_all > t1d) & (time_all < t1d + swin)

    else:
        mask_total = time_all > tge
        mask_singlet = (time_all > tge) & (time_all < tge + swin)

    pe_singlet = ak.to_numpy(
        ak.fill_none(ak.nansum(pe_all[mask_singlet], axis=-1), 0), allow_missing=False
    )
    pe_total = ak.to_numpy(
        ak.fill_none(ak.nansum(pe_all[mask_total], axis=-1), 0), allow_missing=False
    )
    etc = np.divide(
        pe_singlet, pe_total, out=np.full_like(pe_total, np.nan), where=pe_total != 0
    )

    return Array(nda=etc)


# returns relative time shift of the first LAr pulse relative to the Ge trigger
def get_time_shift(
    f_hit,
    f_dsp,
    f_tcm,
    hit_group,
    dsp_group,
    tcm_group,
    tcm_id_table_pattern,
    chs,
    lim,
    trgr,
    tdefault,
    tmin,
    tmax,
) -> Array:
    store = LH5Store()
    # load TCM data to define an event
    ids = store.read(f"/{tcm_group}/array_id", f_tcm)[0].view_as("np")
    idx = store.read(f"/{tcm_group}/array_idx", f_tcm)[0].view_as("np")
    time_all = ak.Array([[] for x in range(np.max(idx) + 1)])

    if isinstance(trgr, (float, int)):
        tge = cast_trigger(trgr, tdefault, length=np.max(idx) + 1)
    else:
        tge = cast_trigger(trgr, tdefault, length=None)

    for ch in chs:
        idx_ch = idx[ids == utils.get_tcm_id_by_pattern(tcm_id_table_pattern, ch)]

        pe = store.read(f"{ch}/{hit_group}/energy_in_pe", f_hit, idx=idx_ch)[0].view_as(
            "np"
        )
        tmp = np.full((np.max(idx) + 1, len(pe[0])), np.nan)
        tmp[idx_ch] = pe
        pe = ak.drop_none(ak.nan_to_none(ak.Array(tmp)))

        # times are in sample units
        times = store.read(f"{ch}/{hit_group}/trigger_pos", f_hit, idx=idx_ch)[
            0
        ].view_as("np")
        tmp = np.full((np.max(idx) + 1, len(times[0])), np.nan)
        tmp[idx_ch] = times
        times = ak.drop_none(ak.nan_to_none(ak.Array(tmp)))

        mask = get_spm_mask(lim, tge, tmin, tmax, pe, times)

        # apply mask and convert sample units to ns
        times = times[mask] * 16

        time_all = ak.concatenate((time_all, times), axis=-1)

    out = ak.min(time_all, axis=-1)

    # Convert to 1D numpy array
    out = ak.to_numpy(ak.fill_none(out, np.inf), allow_missing=False)
    tge = ak.to_numpy(tge, allow_missing=False)

    return Array(out - tge)



from __future__ import annotations

import logging
from types import FunctionType

import numpy as np
import pandas as pd
from iminuit import Minuit, cost, util
from lgdo import Table, lh5

log = logging.getLogger(__name__)
sto = lh5.LH5Store()


def return_nans(input):
    if isinstance(input, FunctionType):
        args = input.__code__.co_varnames[: input.__code__.co_argcount][1:]
        c = cost.UnbinnedNLL(np.array([0]), input)
        m = Minuit(c, *[np.nan for arg in args])
        return m.values, m.errors, np.full((len(m.values), len(m.values)), np.nan)
    else:
        args = input.pdf.__code__.co_varnames[: input.pdf.__code__.co_argcount][1:]
        c = cost.UnbinnedNLL(np.array([0]), input.pdf)
        m = Minuit(c, *[np.nan for arg in args])
        return m.values, m.errors, np.full((len(m.values), len(m.values)), np.nan)


def get_params(file_params, param_list):
    out_params = []
    if isinstance(file_params, dict):
        possible_keys = file_params.keys()
    elif isinstance(file_params, list):
        possible_keys = file_params
    for param in param_list:
        for key in possible_keys:
            if key in param:
                out_params.append(key)
    return np.unique(out_params).tolist()


def load_data(
    files: list,
    lh5_path: str,
    cal_dict: dict,
    params=["cuspEmax"],
    cal_energy_param: str = "cuspEmax_ctc_cal",
    threshold=None,
    return_selection_mask=False,
) -> tuple(np.array, np.array, np.array, np.array):
    """
    Loads in the A/E parameters needed and applies calibration constants to energy
    """

    out_df = pd.DataFrame(columns=params)

    if isinstance(files, dict):
        keys = lh5.ls(
            files[list(files)[0]][0],
            lh5_path if lh5_path[-1] == "/" else lh5_path + "/",
        )
        keys = [key.split("/")[-1] for key in keys]
        if list(files)[0] in cal_dict:
            params = get_params(keys + list(cal_dict[list(files)[0]].keys()), params)
        else:
            params = get_params(keys + list(cal_dict.keys()), params)

        df = []
        all_files = []
        masks = np.array([], dtype=bool)
        for tstamp, tfiles in files.items():
            table = sto.read(lh5_path, tfiles)[0]

            file_df = pd.DataFrame(columns=params)
            if tstamp in cal_dict:
                cal_dict_ts = cal_dict[tstamp]
            else:
                cal_dict_ts = cal_dict

            for outname, info in cal_dict_ts.items():
                outcol = table.eval(info["expression"], info.get("parameters", None))
                table.add_column(outname, outcol)

            for param in params:
                file_df[param] = table[param]

            file_df["run_timestamp"] = np.full(len(file_df), tstamp, dtype=object)

            if threshold is not None:
                mask = file_df[cal_energy_param] > threshold
                file_df.drop(np.where(~mask)[0], inplace=True)
            else:
                mask = np.ones(len(file_df), dtype=bool)
            masks = np.append(masks, mask)
            df.append(file_df)
            all_files += tfiles

        params.append("run_timestamp")
        df = pd.concat(df)

    elif isinstance(files, list):
        keys = lh5.ls(files[0], lh5_path if lh5_path[-1] == "/" else lh5_path + "/")
        keys = [key.split("/")[-1] for key in keys]
        params = get_params(keys + list(cal_dict.keys()), params)

        table = sto.read(lh5_path, files)[0]
        df = pd.DataFrame(columns=params)
        for outname, info in cal_dict.items():
            outcol = table.eval(info["expression"], info.get("parameters", None))
            table.add_column(outname, outcol)
        for param in params:
            df[param] = table[param]
        if threshold is not None:
            masks = df[cal_energy_param] > threshold
            df.drop(np.where(~masks)[0], inplace=True)
        else:
            masks = np.ones(len(df), dtype=bool)
        all_files = files

    for col in list(df.keys()):
        if col not in params:
            df.drop(col, inplace=True, axis=1)

    log.debug(f"data loaded")
    if return_selection_mask:
        return df, masks
    else:
        return df


def get_tcm_pulser_ids(tcm_file, channel, multiplicity_threshold):
    if isinstance(channel, str):
        if channel[:2] == "ch":
            chan = int(channel[2:])
        else:
            chan = int(channel)
    else:
        chan = channel
    if isinstance(tcm_file, list):
        mask = np.array([], dtype=bool)
        for file in tcm_file:
            _, file_mask = get_tcm_pulser_ids(file, chan, multiplicity_threshold)
            mask = np.append(mask, file_mask)
        ids = np.where(mask)[0]
    else:
        data = pd.DataFrame(
            {
                "array_id": sto.read("hardware_tcm_1/array_id", tcm_file)[0].view_as(
                    "np"
                ),
                "array_idx": sto.read("hardware_tcm_1/array_idx", tcm_file)[0].view_as(
                    "np"
                ),
            }
        )
        cumulength = sto.read("hardware_tcm_1/cumulative_length", tcm_file)[0].view_as(
            "np"
        )
        cumulength = np.append(np.array([0]), cumulength)
        n_channels = np.diff(cumulength)
        evt_numbers = np.repeat(np.arange(0, len(cumulength) - 1), np.diff(cumulength))
        evt_mult = np.repeat(np.diff(cumulength), np.diff(cumulength))
        data["evt_number"] = evt_numbers
        data["evt_mult"] = evt_mult
        high_mult_events = np.where(n_channels > multiplicity_threshold)[0]

        ids = data.query(f"array_id=={channel} and evt_number in @high_mult_events")[
            "array_idx"
        ].to_numpy()
        mask = np.zeros(len(data.query(f"array_id=={channel}")), dtype="bool")
        mask[ids] = True
    return ids, mask



"""
This module provides routines for calculating and applying quality cuts
"""

from __future__ import annotations

import glob
import json
import logging
import os

import lgdo.lh5 as lh5
import numpy as np
import pandas as pd
from lgdo.types import Table
from scipy import stats

import pygama.math.histogram as pgh
import pygama.math.peak_fitting as pgf
import pygama.pargen.energy_cal as pgc

log = logging.getLogger(__name__)


def get_keys(in_data, parameters):
    out_params = []
    if isinstance(in_data, dict):
        possible_keys = in_data.keys()
    elif isinstance(in_data, list):
        possible_keys = in_data
    for param in parameters:
        for key in possible_keys:
            if key in param:
                out_params.append(key)
    return np.unique(out_params).tolist()


def generate_cuts(
    data: dict[str, np.ndarray], parameters: dict[str, int], rounding: int = 4
) -> dict:
    """
    Finds double sided cut boundaries for a file for the parameters specified

    Parameters
    ----------
    data : lh5 table or dictionary of arrays
                data to calculate cuts on
    parameters : dict
                 dictionary with the parameter to be cut and the number of sigmas to cut at
    """

    output_dict = {}
    if isinstance(data, pd.DataFrame):
        pass
    elif isinstance(data, Table):
        data = {entry: data[entry].nda for entry in get_keys(data, parameters)}
        data = pd.DataFrame.from_dict(data)
    elif isinstance(data, dict):
        data = pd.DataFrame.from_dict(data)
    for par in parameters.keys():
        if isinstance(parameters[par], dict):
            if "Lower Boundary" in list(parameters[par]) or "Upper Boundary" in list(
                parameters[par]
            ):
                output_dict[par] = parameters[par].copy()
                if "Lower Boundary" not in parameters[par]:
                    output_dict[par]["Lower Boundary"] = -np.inf
                if "Upper Boundary" not in parameters[par]:
                    output_dict[par]["Upper Boundary"] = np.inf
                continue
        num_sigmas = parameters[par]
        try:
            all_par_array = data[par].to_numpy()
        except KeyError:
            all_par_array = data.eval(par).to_numpy()
        idxs = (all_par_array > np.nanpercentile(all_par_array, 1)) & (
            all_par_array < np.nanpercentile(all_par_array, 99)
        )
        par_array = all_par_array[idxs]
        bin_width = (
            np.nanpercentile(par_array, 70) - np.nanpercentile(par_array, 50)
        ) / 5

        counts, start_bins, var = pgh.get_hist(
            par_array, range=(np.nanmin(par_array), np.nanmax(par_array)), dx=bin_width
        )
        max_idx = np.argmax(counts)
        mu = start_bins[max_idx]
        try:
            fwhm = pgh.get_fwhm(counts, start_bins)[0]
            guess_sig = fwhm / 2.355

            lower_bound = mu - 10 * guess_sig

            upper_bound = mu + 10 * guess_sig

        except:
            bin_range = 1000

            if max_idx < bin_range:
                lower_bound_idx = 0
            else:
                lower_bound_idx = max_idx - bin_range
            lower_bound = start_bins[lower_bound_idx]

            if max_idx > len(start_bins) - bin_range:
                upper_bound_idx = -1
            else:
                upper_bound_idx = max_idx + bin_range

            upper_bound = start_bins[upper_bound_idx]

        if (lower_bound < np.nanmin(par_array)) or (lower_bound > np.nanmax(par_array)):
            lower_bound = np.nanmin(par_array)
        if (upper_bound > np.nanmax(par_array)) or (upper_bound < np.nanmin(par_array)):
            upper_bound = np.nanmax(par_array)

        try:
            counts, bins, var = pgh.get_hist(
                par_array, bins=200, range=(lower_bound, upper_bound)
            )

            bin_centres = pgh.get_bin_centers(bins)

            fwhm = pgh.get_fwhm(counts, bins)[0]
            mean = float(bin_centres[np.argmax(counts)])
            pars, cov = pgf.gauss_mode_width_max(
                counts,
                bins,
                mode_guess=mean,
                n_bins=20,
                cost_func="Least Squares",
                inflate_errors=False,
                gof_method="var",
            )
            mean = pars[0]
            std = fwhm / 2.355

            if mean < np.nanmin(bins) or mean > np.nanmax(bins):
                raise IndexError
        except IndexError:
            bin_range = 5000

            if max_idx < bin_range:
                lower_bound_idx = 0
            else:
                lower_bound_idx = max_idx - bin_range
            lower_bound = start_bins[lower_bound_idx]

            if max_idx > len(start_bins) - bin_range:
                upper_bound_idx = -1
            else:
                upper_bound_idx = max_idx + bin_range
            upper_bound = start_bins[upper_bound_idx]
            counts, bins, var = pgh.get_hist(
                par_array, bins=200, range=(lower_bound, upper_bound)
            )

            bin_centres = pgh.get_bin_centers(bins)

            fwhm = pgh.get_fwhm(counts, bins)[0]
            mean = float(bin_centres[np.argmax(counts)])
            std = fwhm / 2.355

        if isinstance(num_sigmas, (int, float)):
            num_sigmas_left = num_sigmas
            num_sigmas_right = num_sigmas
        elif isinstance(num_sigmas, dict):
            if "left" in num_sigmas:
                num_sigmas_left = num_sigmas["left"]
            else:
                num_sigmas["left"] = np.inf
                num_sigmas_left = np.inf
            if "right" in num_sigmas:
                num_sigmas_right = num_sigmas["right"]
            else:
                num_sigmas["right"] = np.inf
                num_sigmas_right = np.inf
        upper = float((num_sigmas_right * std) + mean)
        lower = float((-num_sigmas_left * std) + mean)
        output_dict[par] = {
            "Mean Value": round(mean, rounding),
            "Sigmas Cut": num_sigmas,
            "Upper Boundary": round(upper, rounding),
            "Lower Boundary": round(lower, rounding),
        }
    return output_dict


def get_cut_indexes(
    all_data: dict[str, np.ndarray], cut_dict: dict, energy_param: str = "trapTmax"
) -> list[int]:
    """
    Returns a mask of the data, for a single file, that passes cuts based on dictionary of cuts
    in form of cut boundaries above
    Parameters
    ----------
    File : dict or lh5_table
           dictionary of parameters + array such as load_nda or lh5 table of params
    Cut_dict : string
                Dictionary file with cuts
    """

    indexes = None
    keys = cut_dict.keys()
    if isinstance(all_data, pd.DataFrame):
        pass
    elif isinstance(all_data, Table):
        cut_keys = list(cut_dict)
        cut_keys.append(energy_param)
        all_data = {
            entry: all_data[entry].nda for entry in get_keys(all_data, cut_keys)
        }
        all_data = pd.DataFrame.from_dict(all_data)
    elif isinstance(all_data, dict):
        all_data = pd.DataFrame.from_dict(all_data)
    for cut in keys:
        try:
            data = all_data[cut]
        except KeyError:
            data = all_data.eval(cut).to_numpy()
        upper = cut_dict[cut]["Upper Boundary"]
        lower = cut_dict[cut]["Lower Boundary"]
        idxs = (data < upper) & (data > lower) & (~np.isnan(data))
        percent = 100 * len(np.where(idxs)[0]) / len(idxs)
        log.info(f"{percent:.2f}% passed {cut} cut")

        # Combine masks
        if indexes is not None:
            indexes = indexes & idxs

        else:
            indexes = idxs
        log.debug(f"{cut} loaded")
    percent = 100 * len(np.where(indexes)[0]) / len(indexes)
    log.info(f"{percent:.2f}% passed all cuts")
    return indexes


def cut_dict_to_hit_dict(cut_dict, final_cut_field="is_valid_cal"):
    out_dict = {}
    symbols = "/-+*"
    replacewith = "_"
    for i, param in enumerate(cut_dict):
        out_dict[
            f"{''.join(replacewith  if c in symbols else c for c in param).replace('(','').replace(')','')}_cut"
        ] = {
            "expression": f"(a<({param}))&(({param})<b)",
            "parameters": {
                "a": cut_dict[param]["Lower Boundary"],
                "b": cut_dict[param]["Upper Boundary"],
            },
        }
    quality_cut_exp = ""
    for par in list(cut_dict)[:-1]:
        quality_cut_exp += f"({''.join(replacewith  if c in symbols else c for c in par).replace('(','').replace(')','')}_cut)&"
    quality_cut_exp += f"({''.join(replacewith  if c in symbols else c for c in list(cut_dict)[-1]).replace('(','').replace(')','')}_cut)"
    out_dict[final_cut_field] = {"expression": quality_cut_exp, "parameters": {}}
    return out_dict


def find_pulser_properties(df, energy="daqenergy"):
    if np.nanmax(df[energy]) > 8000:
        hist, bins, var = pgh.get_hist(
            df[energy], dx=1, range=(1000, np.nanmax(df[energy]))
        )
        allowed_err = 200
    else:
        hist, bins, var = pgh.get_hist(
            df[energy], dx=0.2, range=(500, np.nanmax(df[energy]))
        )
        allowed_err = 50
    if np.any(var == 0):
        var[np.where(var == 0)] = 1
    imaxes = pgc.get_i_local_maxima(hist / np.sqrt(var), 3)
    peak_energies = pgh.get_bin_centers(bins)[imaxes]
    pt_pars, pt_covs = pgc.hpge_fit_E_peak_tops(
        hist, bins, var, peak_energies, n_to_fit=10
    )
    peak_e_err = pt_pars[:, 1] * 4

    allowed_mask = np.ones(len(peak_energies), dtype=bool)
    for i, e in enumerate(peak_energies[1:-1]):
        i += 1
        if peak_e_err[i] > allowed_err:
            continue
        if i == 1:
            if (
                e - peak_e_err[i] < peak_energies[i - 1] + peak_e_err[i - 1]
                and peak_e_err[i - 1] < allowed_err
            ):
                overlap = (
                    peak_energies[i - 1]
                    + peak_e_err[i - 1]
                    - (peak_energies[i] - peak_e_err[i])
                )
                peak_e_err[i] -= overlap * (
                    peak_e_err[i] / (peak_e_err[i] + peak_e_err[i - 1])
                )
                peak_e_err[i - 1] -= overlap * (
                    peak_e_err[i - 1] / (peak_e_err[i] + peak_e_err[i - 1])
                )

        if (
            e + peak_e_err[i] > peak_energies[i + 1] - peak_e_err[i + 1]
            and peak_e_err[i + 1] < allowed_err
        ):
            overlap = (e + peak_e_err[i]) - (peak_energies[i + 1] - peak_e_err[i + 1])
            total = peak_e_err[i] + peak_e_err[i + 1]
            peak_e_err[i] -= (overlap) * (peak_e_err[i] / total)
            peak_e_err[i + 1] -= (overlap) * (peak_e_err[i + 1] / total)

    out_pulsers = []
    for i, e in enumerate(peak_energies[allowed_mask]):
        if peak_e_err[i] > allowed_err:
            continue

        try:
            e_cut = (df[energy] > e - peak_e_err[i]) & (df[energy] < e + peak_e_err[i])
            df_peak = df[e_cut]

            time_since_last = (
                df_peak.timestamp.values[1:] - df_peak.timestamp.values[:-1]
            )

            tsl = time_since_last[
                (time_since_last >= 0)
                & (time_since_last < np.percentile(time_since_last, 99.9))
            ]

            bins = np.arange(0.1, 5, 0.001)
            bcs = pgh.get_bin_centers(bins)
            hist, bins, var = pgh.get_hist(tsl, bins=bins)

            maxs = pgc.get_i_local_maxima(hist, 45)
            maxs = maxs[maxs > 20]

            super_max = pgc.get_i_local_maxima(hist, 500)
            super_max = super_max[super_max > 20]
            if len(maxs) < 2:
                continue
            else:
                max_locs = np.array([0.0])
                max_locs = np.append(max_locs, bcs[np.array(maxs)])
                if (
                    len(np.where(np.abs(np.diff(np.diff(max_locs))) <= 0.001)[0]) > 1
                    or (np.abs(np.diff(np.diff(max_locs))) <= 0.001).all()
                    or len(super_max) > 0
                ):
                    pulser_e = e
                    period = stats.mode(tsl).mode[0]
                    if period > 0.1:
                        out_pulsers.append((pulser_e, peak_e_err[i], period, energy))

                else:
                    continue
        except:
            continue
    return out_pulsers


def tag_pulsers(df, chan_info, window=0.01):
    df["isPulser"] = 0

    if isinstance(chan_info, tuple):
        chan_info = [chan_info]
    final_mask = None
    for chan_i in chan_info:
        pulser_energy, peak_e_err, period, energy_name = chan_i

        e_cut = (df[energy_name] < pulser_energy + peak_e_err) & (
            df[energy_name] > pulser_energy - peak_e_err
        )
        df_pulser = df[e_cut]

        time_since_last = np.zeros(len(df_pulser))
        time_since_last[1:] = (
            df_pulser.timestamp.values[1:] - df_pulser.timestamp.values[:-1]
        )

        mode_idxs = (time_since_last > period - window) & (
            time_since_last < period + window
        )

        pulser_events = np.count_nonzero(mode_idxs)
        # print(f"pulser events: {pulser_events}")
        if pulser_events < 3:
            return df
        df_pulser = df_pulser[mode_idxs]

        ts = df_pulser.timestamp.values
        diff_zero = np.zeros(len(ts))
        diff_zero[1:] = np.around(np.divide(np.subtract(ts[1:], ts[:-1]), period))
        diff_cum = np.cumsum(diff_zero)
        z = np.polyfit(diff_cum, ts, 1)
        p = np.poly1d(z)

        period = z[0]
        phase = z[1]
        pulser_mod = np.abs(df_pulser.timestamp - phase) % period
        mod = np.abs(df.timestamp - phase) % period

        period_cut = (mod < 0.1) | ((period - mod) < 0.1)  # 0.1)

        if final_mask is None:
            final_mask = e_cut & period_cut
        else:
            final_mask = final_mask | (e_cut & period_cut)

    df.loc[final_mask, "isPulser"] = 1

    return df



import os
from pathlib import Path

import awkward as ak
import numpy as np
from lgdo.lh5 import LH5Store

from pygama.evt import build_evt
from pygama.skm import build_skm

config_dir = Path(__file__).parent / "configs"
evt_config_dir = Path(__file__).parent.parent / "evt" / "configs"
store = LH5Store()


def test_basics(lgnd_test_data, tmptestdir):
    outfile = f"{tmptestdir}/l200-p03-r001-phy-20230322T160139Z-tier_evt.lh5"
    tcm_path = "lh5/prod-ref-l200/generated/tier/tcm/phy/p03/r001/l200-p03-r001-phy-20230322T160139Z-tier_tcm.lh5"
    if os.path.exists(outfile):
        os.remove(outfile)
    build_evt(
        f_tcm=lgnd_test_data.get_path(tcm_path),
        f_dsp=lgnd_test_data.get_path(tcm_path.replace("tcm", "dsp")),
        f_hit=lgnd_test_data.get_path(tcm_path.replace("tcm", "hit")),
        f_evt=outfile,
        evt_config=f"{evt_config_dir}/vov-test-evt-config.json",
        wo_mode="o",
        evt_group="evt",
        hit_group="hit",
        dsp_group="dsp",
        tcm_group="hardware_tcm_1",
    )

    skm_conf = f"{config_dir}/basic-skm-config.json"
    skm_out = f"{tmptestdir}/l200-p03-r001-phy-20230322T160139Z-tier_skm.lh5"
    build_skm(
        outfile,
        lgnd_test_data.get_path(tcm_path.replace("tcm", "hit")),
        lgnd_test_data.get_path(tcm_path.replace("tcm", "dsp")),
        lgnd_test_data.get_path(tcm_path),
        skm_out,
        skm_conf,
        wo_mode="o",
    )

    assert os.path.exists(skm_out)
    df = store.read("/skm/", skm_out)[0].view_as("pd")
    assert "timestamp" in df.keys()
    assert "energy_0" in df.keys()
    assert "energy_1" in df.keys()
    assert "energy_2" in df.keys()
    assert "energy_id_0" in df.keys()
    assert "energy_id_1" in df.keys()
    assert "energy_id_2" in df.keys()
    assert "multiplicity" in df.keys()
    assert "energy_sum" in df.keys()
    assert (df.multiplicity.to_numpy() <= 3).all()
    assert (
        np.nan_to_num(df.energy_0.to_numpy())
        + np.nan_to_num(df.energy_1.to_numpy())
        + np.nan_to_num(df.energy_2.to_numpy())
        == df.energy_sum.to_numpy()
    ).all()

    vov_eid = ak.to_numpy(
        ak.fill_none(
            ak.pad_none(
                store.read("/evt/energy_id", outfile)[0].view_as("ak"), 3, clip=True
            ),
            0,
        ),
        allow_missing=False,
    )
    assert (vov_eid[:, 0] == df.energy_id_0.to_numpy()).all()
    assert (vov_eid[:, 1] == df.energy_id_1.to_numpy()).all()
    assert (vov_eid[:, 2] == df.energy_id_2.to_numpy()).all()


def test_attribute_passing(lgnd_test_data, tmptestdir):
    outfile = f"{tmptestdir}/l200-p03-r001-phy-20230322T160139Z-tier_evt.lh5"
    tcm_path = "lh5/prod-ref-l200/generated/tier/tcm/phy/p03/r001/l200-p03-r001-phy-20230322T160139Z-tier_tcm.lh5"
    if os.path.exists(outfile):
        os.remove(outfile)
    build_evt(
        f_tcm=lgnd_test_data.get_path(tcm_path),
        f_dsp=lgnd_test_data.get_path(tcm_path.replace("tcm", "dsp")),
        f_hit=lgnd_test_data.get_path(tcm_path.replace("tcm", "hit")),
        f_evt=outfile,
        evt_config=f"{evt_config_dir}/vov-test-evt-config.json",
        wo_mode="o",
        evt_group="evt",
        hit_group="hit",
        dsp_group="dsp",
        tcm_group="hardware_tcm_1",
    )

    skm_conf = f"{config_dir}/basic-skm-config.json"

    skm_out = f"{tmptestdir}/l200-p03-r001-phy-20230322T160139Z-tier_skm.lh5"

    build_skm(
        outfile,
        lgnd_test_data.get_path(tcm_path.replace("tcm", "hit")),
        lgnd_test_data.get_path(tcm_path.replace("tcm", "dsp")),
        lgnd_test_data.get_path(tcm_path),
        skm_out,
        skm_conf,
        wo_mode="o",
    )

    assert os.path.exists(skm_out)
    assert "info" in store.read("/skm/timestamp", skm_out)[0].getattrs().keys()
    assert store.read("/skm/timestamp", skm_out)[0].getattrs()["info"] == "pk was here"



"""
Module provides LEGEND internal functions
"""
from importlib import import_module

from lgdo.lh5 import utils


def metadata(params: dict) -> list:
    # only import legend meta data when needed.
    # LEGEND collaborators can use the meta keyword
    # While for users w/o access to the LEGEND meta data this is still working
    lm = import_module("legendmeta")
    lmeta = lm.LegendMetadata(path=utils.expand_path(params["meta_path"]))
    chmap = lmeta.channelmap(params["time_key"])

    tmp = [
        f"ch{e}"
        for e in chmap.map("daq.rawid")
        if chmap.map("daq.rawid")[e]["system"] == params["system"]
    ]

    if "selectors" in params.keys():
        for k in params["selectors"].keys():
            s = ""
            for e in k.split("."):
                s += f"['{e}']"

            tmp = [
                e
                for e in tmp
                if eval("dotter" + s, {"dotter": chmap.map("daq.rawid")[int(e[2:])]})
                == params["selectors"][k]
            ]
    return tmp



"""
This module provides utilities to build the `evt` tier.
"""

from __future__ import annotations

import re

import awkward as ak
import numpy as np
from lgdo.lh5 import LH5Store
from numpy.typing import NDArray


def get_tcm_id_by_pattern(tcm_id_table_pattern: str, ch: str) -> int:
    pre = tcm_id_table_pattern.split("{")[0]
    post = tcm_id_table_pattern.split("}")[1]
    return int(ch.strip(pre).strip(post))


def get_table_name_by_pattern(tcm_id_table_pattern: str, ch_id: int) -> str:
    # check tcm_id_table_pattern validity
    pattern_check = re.findall(r"{([^}]*?)}", tcm_id_table_pattern)[0]
    if pattern_check == "" or ":" == pattern_check[0]:
        return tcm_id_table_pattern.format(ch_id)
    else:
        raise NotImplementedError(
            "Only empty placeholders with format specifications are currently implemented"
        )


def num_and_pars(value: str, par_dic: dict):
    # function tries to convert a string to a int, float, bool
    # or returns the value if value is a key in par_dic
    if value in par_dic.keys():
        return par_dic[value]
    try:
        value = int(value)
    except ValueError:
        try:
            value = float(value)
        except ValueError:
            try:
                value = bool(value)
            except ValueError:
                pass
    return value


def find_parameters(
    f_hit: str,
    f_dsp: str,
    ch: str,
    idx_ch: NDArray,
    exprl: list,
    hit_group: str = "hit",
    dsp_group: str = "dsp",
) -> dict:
    """Wraps :func:`load_vars_to_nda` to return parameters from `hit` and `dsp`
    tiers.

    Parameters
    ----------
    f_hit
       path to `hit` tier file.
    f_dsp
       path to `dsp` tier file.
    ch
       "rawid" in the tiers.
    idx_ch
       index array of entries to be read from files.
    exprl
       list of tuples ``(tier, field)`` to be found in the `hit/dsp` tiers.
    dsp_group
        LH5 root group in dsp file.
    hit_group
        LH5 root group in hit file.
    """

    # find fields in either dsp, hit
    dsp_flds = [e[1] for e in exprl if e[0] == dsp_group]
    hit_flds = [e[1] for e in exprl if e[0] == hit_group]

    store = LH5Store()
    hit_dict, dsp_dict = {}, {}
    if len(hit_flds) > 0:
        hit_ak = store.read(
            f"{ch.replace('/','')}/{hit_group}/", f_hit, field_mask=hit_flds, idx=idx_ch
        )[0].view_as("ak")
        hit_dict = dict(
            zip([f"{hit_group}_" + e for e in ak.fields(hit_ak)], ak.unzip(hit_ak))
        )
    if len(dsp_flds) > 0:
        dsp_ak = store.read(
            f"{ch.replace('/','')}/{dsp_group}/", f_dsp, field_mask=dsp_flds, idx=idx_ch
        )[0].view_as("ak")
        dsp_dict = dict(
            zip([f"{dsp_group}_" + e for e in ak.fields(dsp_ak)], ak.unzip(dsp_ak))
        )

    return hit_dict | dsp_dict


def get_data_at_channel(
    ch: str,
    ids: NDArray,
    idx: NDArray,
    expr: str,
    exprl: list,
    var_ph: dict,
    is_evaluated: bool,
    f_hit: str,
    f_dsp: str,
    defv,
    tcm_id_table_pattern: str = "ch{}",
    evt_group: str = "evt",
    hit_group: str = "hit",
    dsp_group: str = "dsp",
) -> np.ndarray:
    """Evaluates an expression and returns the result.

    Parameters
    ----------
    ch
       "rawid" of channel to be evaluated.
    idx
       `tcm` index array.
    ids
       `tcm` id array.
    expr
       expression to be evaluated.
    exprl
       list of parameter-tuples ``(root_group, field)`` found in the expression.
    var_ph
       dict of additional parameters that are not channel dependent.
    is_evaluated
       if false, the expression does not get evaluated but an array of default
       values is returned.
    f_hit
       path to `hit` tier file.
    f_dsp
       path to `dsp` tier file.
    defv
       default value.
    tcm_id_table_pattern
        Pattern to format tcm id values to table name in higher tiers. Must have one
        placeholder which is the tcm id.
    dsp_group
        LH5 root group in dsp file.
    hit_group
        LH5 root group in hit file.
    evt_group
        LH5 root group in evt file.
    """

    # get index list for this channel to be loaded
    idx_ch = idx[ids == get_tcm_id_by_pattern(tcm_id_table_pattern, ch)]
    outsize = len(idx_ch)

    if not is_evaluated:
        res = np.full(outsize, defv, dtype=type(defv))
    elif "tcm.array_id" == expr:
        res = np.full(
            outsize, get_tcm_id_by_pattern(tcm_id_table_pattern, ch), dtype=int
        )
    elif "tcm.index" == expr:
        res = np.where(ids == get_tcm_id_by_pattern(tcm_id_table_pattern, ch))[0]
    else:
        var = find_parameters(
            f_hit=f_hit,
            f_dsp=f_dsp,
            ch=ch,
            idx_ch=idx_ch,
            exprl=exprl,
            hit_group=hit_group,
            dsp_group=dsp_group,
        )

        if var_ph is not None:
            var = var | var_ph

        # evaluate expression
        # move tier+dots in expression to underscores (e.g. evt.foo -> evt_foo)
        res = eval(
            expr.replace(f"{dsp_group}.", f"{dsp_group}_")
            .replace(f"{hit_group}.", f"{hit_group}_")
            .replace(f"{evt_group}.", ""),
            var,
        )

        # in case the expression evaluates to a single value blow it up
        if (not hasattr(res, "__len__")) or (isinstance(res, str)):
            return np.full(outsize, res)

        # the resulting arrays need to be 1D from the operation,
        # this can only change once we support larger than two dimensional LGDOs
        # ak.to_numpy() raises error if array not regular
        res = ak.to_numpy(res, allow_missing=False)

        # in this method only 1D values are allowed
        if res.ndim > 1:
            raise ValueError(
                f"expression '{expr}' must return 1D array. If you are using VectorOfVectors or ArrayOfEqualSizedArrays, use awkward reduction functions to reduce the dimension"
            )

    return res


def get_mask_from_query(
    qry: str | NDArray,
    length: int,
    ch: str,
    idx_ch: NDArray,
    f_hit: str,
    f_dsp: str,
    hit_group: str = "hit",
    dsp_group: str = "dsp",
) -> np.ndarray:
    """Evaluates a query expression and returns a mask accordingly.

    Parameters
    ----------
    qry
       query expression.
    length
       length of the return mask.
    ch
       "rawid" of channel to be evaluated.
    idx_ch
       channel indices to be read.
    f_hit
       path to `hit` tier file.
    f_dsp
       path to `dsp` tier file.
    hit_group
        LH5 root group in hit file.
    dsp_group
        LH5 root group in dsp file.
    """

    # get sub evt based query condition if needed
    if isinstance(qry, str):
        qry_lst = re.findall(r"(hit|dsp).([a-zA-Z_$][\w$]*)", qry)
        qry_var = find_parameters(
            f_hit=f_hit,
            f_dsp=f_dsp,
            ch=ch,
            idx_ch=idx_ch,
            exprl=qry_lst,
            hit_group=hit_group,
            dsp_group=dsp_group,
        )
        limarr = eval(
            qry.replace(f"{dsp_group}.", f"{dsp_group}_").replace(
                f"{hit_group}.", f"{hit_group}_"
            ),
            qry_var,
        )

        # in case the expression evaluates to a single value blow it up
        if (not hasattr(limarr, "__len__")) or (isinstance(limarr, str)):
            return np.full(len(idx_ch), limarr)

        limarr = ak.to_numpy(limarr, allow_missing=False)
        if limarr.ndim > 1:
            raise ValueError(
                f"query '{qry}' must return 1D array. If you are using VectorOfVectors or ArrayOfEqualSizedArrays, use awkward reduction functions to reduce the dimension"
            )

    # or forward the array
    elif isinstance(qry, np.ndarray):
        limarr = qry

    # if no condition, it must be true
    else:
        limarr = np.ones(length).astype(bool)

    # explicit cast to bool
    if limarr.dtype != bool:
        limarr = limarr.astype(bool)

    return limarr



"""
Contains submodules for evt processing
"""

from .spm import (
    get_energy,
    get_energy_dplms,
    get_etc,
    get_majority,
    get_majority_dplms,
    get_time_shift,
)

__all__ = [
    "get_energy",
    "get_majority",
    "get_energy_dplms",
    "get_majority_dplms",
    "get_etc",
    "get_time_shift",
]



"""
This module implements routines to build the `evt` tier.
"""

from __future__ import annotations

import itertools
import json
import logging
import re
from importlib import import_module

import awkward as ak
import numpy as np
from lgdo import Array, ArrayOfEqualSizedArrays, Table, VectorOfVectors, lh5
from lgdo.lh5 import LH5Store

from . import aggregators, utils

log = logging.getLogger(__name__)


def build_evt(
    f_tcm: str,
    f_dsp: str,
    f_hit: str,
    f_evt: str,
    evt_config: str | dict,
    wo_mode: str = "write_safe",
    evt_group: str = "evt",
    tcm_group: str = "hardware_tcm_1",
    dsp_group: str = "dsp",
    hit_group: str = "hit",
    tcm_id_table_pattern: str = "ch{}",
) -> None:
    """Transform data from the `hit` and `dsp` levels which a channel sorted to a
    event sorted data format.

    Parameters
    ----------
    f_tcm
        input LH5 file of the `tcm` level.
    f_dsp
        input LH5 file of the `dsp` level.
    f_hit
        input LH5 file of the `hit` level.
    f_evt
        name of the output file.
    evt_config
        name of configuration file or dictionary defining event fields. Channel
        lists can be defined by importing a metadata module.

        - ``operations`` defines the fields ``name=key``, where ``channels``
          specifies the channels used to for this field (either a string or a
          list of strings),
        - ``aggregation_mode`` defines how the channels should be combined (see
          :func:`evaluate_expression`).
        - ``expression`` defnies the mathematical/special function to apply
          (see :func:`evaluate_expression`),
        - ``query`` defines an expression to mask the aggregation.
        - ``parameters`` defines any other parameter used in expression.

        For example:

        .. code-block:: json

            {
              "channels": {
                "geds_on": ["ch1084803", "ch1084804", "ch1121600"],
                "spms_on": ["ch1057600", "ch1059201", "ch1062405"],
                "muon": "ch1027202",
              },
              "operations": {
                "energy_id":{
                  "channels": "geds_on",
                  "aggregation_mode": "gather",
                  "query": "hit.cuspEmax_ctc_cal > 25",
                  "expression": "tcm.array_id",
                  "sort": "ascend_by:dsp.tp_0_est"
                },
                "energy":{
                  "aggregation_mode": "keep_at_ch:evt.energy_id",
                  "expression": "hit.cuspEmax_ctc_cal > 25"
                }
                "is_muon_rejected":{
                  "channels": "muon",
                  "aggregation_mode": "any",
                  "expression": "dsp.wf_max>a",
                  "parameters": {"a":15100},
                  "initial": false
                },
                "multiplicity":{
                  "channels":  ["geds_on", "geds_no_psd", "geds_ac"],
                  "aggregation_mode": "sum",
                  "expression": "hit.cuspEmax_ctc_cal > a",
                  "parameters": {"a":25},
                  "initial": 0
                },
                "t0":{
                  "aggregation_mode": "keep_at_ch:evt.energy_id",
                  "expression": "dsp.tp_0_est"
                },
                "lar_energy":{
                  "channels": "spms_on",
                  "aggregation_mode": "function",
                  "expression": ".modules.spm.get_energy(0.5, evt.t0, 48000, 1000, 5000)"
                },
              }
            }

    wo_mode
        writing mode.
    evt group
        LH5 root group name of `evt` tier.
    tcm_group
        LH5 root group in `tcm` file.
    dsp_group
        LH5 root group in `dsp` file.
    hit_group
        LH5 root group in `hit` file.
    tcm_id_table_pattern
        pattern to format `tcm` id values to table name in higher tiers. Must
        have one placeholder which is the `tcm` id.
    """

    store = LH5Store()
    tbl_cfg = evt_config
    if not isinstance(tbl_cfg, (str, dict)):
        raise TypeError()
    if isinstance(tbl_cfg, str):
        with open(tbl_cfg) as f:
            tbl_cfg = json.load(f)

    if "channels" not in tbl_cfg.keys():
        raise ValueError("channel field needs to be specified in the config")
    if "operations" not in tbl_cfg.keys():
        raise ValueError("operations field needs to be specified in the config")

    # check tcm_id_table_pattern validity
    pattern_check = re.findall(r"{([^}]*?)}", tcm_id_table_pattern)
    if len(pattern_check) != 1:
        raise ValueError(
            f"tcm_id_table_pattern must have exactly one placeholder. {tcm_id_table_pattern} is invalid."
        )
    elif "{" in pattern_check[0] or "}" in pattern_check[0]:
        raise ValueError(
            f"tcm_id_table_pattern {tcm_id_table_pattern} has an invalid placeholder."
        )

    if (
        utils.get_table_name_by_pattern(
            tcm_id_table_pattern,
            utils.get_tcm_id_by_pattern(tcm_id_table_pattern, lh5.ls(f_hit)[0]),
        )
        != lh5.ls(f_hit)[0]
    ):
        raise ValueError(
            f"tcm_id_table_pattern {tcm_id_table_pattern} does not match keys in data!"
        )

    # create channel list according to config
    # This can be either read from the meta data
    # or a list of channel names
    log.debug("Creating channel dictionary")

    chns = {}

    for k, v in tbl_cfg["channels"].items():
        if isinstance(v, dict):
            # it is a meta module. module_name must exist
            if "module" not in v.keys():
                raise ValueError(
                    "Need module_name to load channel via a meta data module"
                )

            attr = {}
            # the time_key argument is set to the time key of the DSP file
            # in case it is not provided by the config
            if "time_key" not in v.keys():
                attr["time_key"] = re.search(r"\d{8}T\d{6}Z", f_dsp).group(0)

            # if "None" do None
            elif "None" == v["time_key"]:
                attr["time_key"] = None

            # load module
            p, m = v["module"].rsplit(".", 1)
            met = getattr(import_module(p, package=__package__), m)
            chns[k] = met(v | attr)

        elif isinstance(v, str):
            chns[k] = [v]

        elif isinstance(v, list):
            chns[k] = [e for e in v]

    nrows = store.read_n_rows(f"/{tcm_group}/cumulative_length", f_tcm)

    table = Table(size=nrows)

    for k, v in tbl_cfg["operations"].items():
        log.debug("Processing field " + k)

        # if mode not defined in operation, it can only be an operation on the evt level.
        if "aggregation_mode" not in v.keys():
            var = {}
            if "parameters" in v.keys():
                var = var | v["parameters"]
            res = table.eval(v["expression"].replace(f"{evt_group}.", ""), var)

            # add attribute if present
            if "lgdo_attrs" in v.keys():
                res.attrs |= v["lgdo_attrs"]

            table.add_field(k, res)

        # Else we build the event entry
        else:
            if "channels" not in v.keys():
                chns_e = []
            elif isinstance(v["channels"], str):
                chns_e = chns[v["channels"]]
            elif isinstance(v["channels"], list):
                chns_e = list(
                    itertools.chain.from_iterable([chns[e] for e in v["channels"]])
                )
            chns_rm = []
            if "exclude_channels" in v.keys():
                if isinstance(v["exclude_channels"], str):
                    chns_rm = chns[v["exclude_channels"]]
                elif isinstance(v["exclude_channels"], list):
                    chns_rm = list(
                        itertools.chain.from_iterable(
                            [chns[e] for e in v["exclude_channels"]]
                        )
                    )

            pars, qry, defaultv, srter = None, None, np.nan, None
            if "parameters" in v.keys():
                pars = v["parameters"]
            if "query" in v.keys():
                qry = v["query"]
            if "initial" in v.keys():
                defaultv = v["initial"]
                if isinstance(defaultv, str) and (
                    defaultv in ["np.nan", "np.inf", "-np.inf"]
                ):
                    defaultv = eval(defaultv)
            if "sort" in v.keys():
                srter = v["sort"]

            obj = evaluate_expression(
                f_tcm=f_tcm,
                f_hit=f_hit,
                f_dsp=f_dsp,
                chns=chns_e,
                chns_rm=chns_rm,
                mode=v["aggregation_mode"],
                expr=v["expression"],
                nrows=nrows,
                table=table,
                para=pars,
                qry=qry,
                defv=defaultv,
                sorter=srter,
                tcm_id_table_pattern=tcm_id_table_pattern,
                evt_group=evt_group,
                hit_group=hit_group,
                dsp_group=dsp_group,
                tcm_group=tcm_group,
            )

            # add attribute if present
            if "lgdo_attrs" in v.keys():
                obj.attrs |= v["lgdo_attrs"]

            table.add_field(k, obj)

    # write output fields into f_evt
    if "outputs" in tbl_cfg.keys():
        if len(tbl_cfg["outputs"]) < 1:
            log.warning("No output fields specified, no file will be written.")
        else:
            clms_to_remove = [e for e in table.keys() if e not in tbl_cfg["outputs"]]
            for fld in clms_to_remove:
                table.remove_field(fld, True)
            store.write(
                obj=table, name=f"/{evt_group}/", lh5_file=f_evt, wo_mode=wo_mode
            )
    else:
        log.warning("No output fields specified, no file will be written.")

    key = re.search(r"\d{8}T\d{6}Z", f_hit).group(0)
    log.info(
        f"Applied {len(tbl_cfg['operations'])} operations to key {key} and saved {len(tbl_cfg['outputs'])} evt fields across {len(chns)} channel groups"
    )


def evaluate_expression(
    f_tcm: str,
    f_hit: str,
    f_dsp: str,
    chns: list,
    chns_rm: list,
    mode: str,
    expr: str,
    nrows: int,
    table: Table = None,
    para: dict = None,
    qry: str = None,
    defv: bool | int | float = np.nan,
    sorter: str = None,
    tcm_id_table_pattern: str = "ch{}",
    evt_group: str = "evt",
    hit_group: str = "hit",
    dsp_group: str = "dsp",
    tcm_group: str = "tcm",
) -> Array | ArrayOfEqualSizedArrays | VectorOfVectors:
    """Evaluates the expression defined by the user across all channels
    according to the mode.

    Parameters
    ----------
    f_tcm
       path to `tcm` tier file.
    f_hit
       path to `hit` tier file.
    f_dsp
       path to `dsp` tier file.
    chns
       list of channel names across which expression gets evaluated (form:
       ``ch<rawid>``).
    chns_rm
       list of channels which get set to default value during evaluation. In
       function mode they are removed entirely (form: ``ch<rawid>``)
    mode
       The mode determines how the event entry is calculated across channels.
       Options are:

       - ``first_at:sorter``: aggregates across channels by returning the
         expression of the channel with smallest value of sorter.
       - ``last_at``: aggregates across channels by returning the expression of
         the channel with largest value of sorter.
       - ``sum``: aggregates by summation.
       - ``any``: aggregates by logical or.
       - ``all``: aggregates by logical and.
       - ``keep_at_ch:ch_field``: aggregates according to passed ch_field.
       - ``keep_at_idx:tcm_idx_field``: aggregates according to passed tcm
         index field.
       - ``gather``: Channels are not combined, but result saved as
         :class:`.VectorOfVectors`.

    qry
       a query that can mask the aggregation.
    expr
       the expression. That can be any mathematical equation/comparison. If
       `mode` is ``function``, the expression needs to be a special processing
       function defined in modules (e.g. :func:`.modules.spm.get_energy`). In
       the expression parameters from either hit, dsp, evt tier (from
       operations performed before this one! Dictionary operations order
       matters), or from the ``parameters`` field can be used.
    nrows
       number of rows to be processed.
    table
       table of `evt` tier data.
    para
       dictionary of parameters defined in the ``parameters`` field in the
       configuration dictionary.
    defv
       default value of evaluation.
    sorter
       can be used to sort vector outputs according to sorter expression (see
       :func:`evaluate_to_vector`).
    tcm_id_table_pattern
        pattern to format tcm id values to table name in higher tiers. Must have one
        placeholder which is the `tcm` id.
    evt group
        LH5 root group name of `evt` tier.
    tcm_group
        LH5 root group in `tcm` file.
    dsp_group
        LH5 root group in `dsp` file.
    hit_group
        LH5 root group in `hit` file.
    """

    store = LH5Store()

    # find parameters in evt file or in parameters
    exprl = re.findall(
        rf"({evt_group}|{hit_group}|{dsp_group}).([a-zA-Z_$][\w$]*)", expr
    )
    var_ph = {}
    if table:
        var_ph = var_ph | {
            e: table[e].view_as("ak")
            for e in table.keys()
            if isinstance(table[e], (Array, ArrayOfEqualSizedArrays, VectorOfVectors))
        }
    if para:
        var_ph = var_ph | para

    if mode == "function":
        # evaluate expression
        func, params = expr.split("(")
        params = (
            params.replace(f"{dsp_group}.", f"{dsp_group}_")
            .replace(f"{hit_group}.", f"{hit_group}_")
            .replace(f"{evt_group}.", "")
        )
        params = [
            f_hit,
            f_dsp,
            f_tcm,
            hit_group,
            dsp_group,
            tcm_group,
            tcm_id_table_pattern,
            [x for x in chns if x not in chns_rm],
        ] + [utils.num_and_pars(e, var_ph) for e in params[:-1].split(",")]

        # load function dynamically
        p, m = func.rsplit(".", 1)
        met = getattr(import_module(p, package=__package__), m)
        return met(*params)

    else:
        # check if query is either on channel basis or evt basis (and not a mix)
        qry_mask = qry
        if qry is not None:
            if f"{evt_group}." in qry and (
                f"{hit_group}." in qry or f"{dsp_group}." in qry
            ):
                raise ValueError(
                    f"Query can't be a mix of {evt_group} tier and lower tiers."
                )

            # if it is an evt query we can evaluate it directly here
            if table and f"{evt_group}." in qry:
                qry_mask = eval(qry.replace(f"{evt_group}.", ""), table)

        # load TCM data to define an event
        ids = store.read(f"/{tcm_group}/array_id", f_tcm)[0].view_as("np")
        idx = store.read(f"/{tcm_group}/array_idx", f_tcm)[0].view_as("np")
        cumulength = store.read(f"/{tcm_group}/cumulative_length", f_tcm)[0].view_as(
            "np"
        )

        # switch through modes
        if table and (("keep_at_ch:" == mode[:11]) or ("keep_at_idx:" == mode[:12])):
            if "keep_at_ch:" == mode[:11]:
                ch_comp = table[mode[11:].replace(f"{evt_group}.", "")]
            else:
                ch_comp = table[mode[12:].replace(f"{evt_group}.", "")]
                if isinstance(ch_comp, Array):
                    ch_comp = Array(nda=ids[ch_comp.view_as("np")])
                elif isinstance(ch_comp, VectorOfVectors):
                    ch_comp = ch_comp.view_as("ak")
                    ch_comp = VectorOfVectors(
                        array=ak.unflatten(
                            ids[ak.flatten(ch_comp)], ak.count(ch_comp, axis=-1)
                        )
                    )
                else:
                    raise NotImplementedError(
                        type(ch_comp)
                        + " not supported (only Array and VectorOfVectors are supported)"
                    )

            if isinstance(ch_comp, Array):
                return aggregators.evaluate_at_channel(
                    cumulength=cumulength,
                    idx=idx,
                    ids=ids,
                    f_hit=f_hit,
                    f_dsp=f_dsp,
                    chns_rm=chns_rm,
                    expr=expr,
                    exprl=exprl,
                    ch_comp=ch_comp,
                    var_ph=var_ph,
                    defv=defv,
                    tcm_id_table_pattern=tcm_id_table_pattern,
                    evt_group=evt_group,
                    hit_group=hit_group,
                    dsp_group=dsp_group,
                )
            elif isinstance(ch_comp, VectorOfVectors):
                return aggregators.evaluate_at_channel_vov(
                    cumulength=cumulength,
                    idx=idx,
                    ids=ids,
                    f_hit=f_hit,
                    f_dsp=f_dsp,
                    expr=expr,
                    exprl=exprl,
                    ch_comp=ch_comp,
                    chns_rm=chns_rm,
                    var_ph=var_ph,
                    defv=defv,
                    tcm_id_table_pattern=tcm_id_table_pattern,
                    evt_group=evt_group,
                    hit_group=hit_group,
                    dsp_group=dsp_group,
                )
            else:
                raise NotImplementedError(
                    type(ch_comp)
                    + " not supported (only Array and VectorOfVectors are supported)"
                )
        elif "first_at:" in mode or "last_at:" in mode:
            sorter = tuple(
                re.findall(
                    rf"({evt_group}|{hit_group}|{dsp_group}).([a-zA-Z_$][\w$]*)",
                    mode.split("first_at:")[-1],
                )[0]
            )
            return aggregators.evaluate_to_first_or_last(
                cumulength=cumulength,
                idx=idx,
                ids=ids,
                f_hit=f_hit,
                f_dsp=f_dsp,
                chns=chns,
                chns_rm=chns_rm,
                expr=expr,
                exprl=exprl,
                qry=qry_mask,
                nrows=nrows,
                sorter=sorter,
                var_ph=var_ph,
                defv=defv,
                is_first=True if "first_at:" in mode else False,
                tcm_id_table_pattern=tcm_id_table_pattern,
                evt_group=evt_group,
                hit_group=hit_group,
                dsp_group=dsp_group,
            )
        elif mode in ["sum", "any", "all"]:
            return aggregators.evaluate_to_scalar(
                mode=mode,
                cumulength=cumulength,
                idx=idx,
                ids=ids,
                f_hit=f_hit,
                f_dsp=f_dsp,
                chns=chns,
                chns_rm=chns_rm,
                expr=expr,
                exprl=exprl,
                qry=qry_mask,
                nrows=nrows,
                var_ph=var_ph,
                defv=defv,
                tcm_id_table_pattern=tcm_id_table_pattern,
                evt_group=evt_group,
                hit_group=hit_group,
                dsp_group=dsp_group,
            )
        elif "gather" == mode:
            return aggregators.evaluate_to_vector(
                cumulength=cumulength,
                idx=idx,
                ids=ids,
                f_hit=f_hit,
                f_dsp=f_dsp,
                chns=chns,
                chns_rm=chns_rm,
                expr=expr,
                exprl=exprl,
                qry=qry_mask,
                nrows=nrows,
                var_ph=var_ph,
                defv=defv,
                sorter=sorter,
                tcm_id_table_pattern=tcm_id_table_pattern,
                evt_group=evt_group,
                hit_group=hit_group,
                dsp_group=dsp_group,
            )
        else:
            raise ValueError(mode + " not a valid mode")



import os
from pathlib import Path

import awkward as ak
import numpy as np
import pytest
from lgdo import Array, VectorOfVectors, lh5
from lgdo.lh5 import LH5Store

from pygama.evt import build_evt

config_dir = Path(__file__).parent / "configs"
store = LH5Store()


def test_basics(lgnd_test_data, tmptestdir):
    outfile = f"{tmptestdir}/l200-p03-r001-phy-20230322T160139Z-tier_evt.lh5"
    tcm_path = "lh5/prod-ref-l200/generated/tier/tcm/phy/p03/r001/l200-p03-r001-phy-20230322T160139Z-tier_tcm.lh5"
    if os.path.exists(outfile):
        os.remove(outfile)
    build_evt(
        f_tcm=lgnd_test_data.get_path(tcm_path),
        f_dsp=lgnd_test_data.get_path(tcm_path.replace("tcm", "dsp")),
        f_hit=lgnd_test_data.get_path(tcm_path.replace("tcm", "hit")),
        f_evt=outfile,
        evt_config=f"{config_dir}/basic-evt-config.json",
        wo_mode="o",
        evt_group="evt",
        hit_group="hit",
        dsp_group="dsp",
        tcm_group="hardware_tcm_1",
    )
    assert "statement" in store.read("/evt/multiplicity", outfile)[0].getattrs().keys()
    assert (
        store.read("/evt/multiplicity", outfile)[0].getattrs()["statement"]
        == "0bb decay is real"
    )
    assert os.path.exists(outfile)
    assert len(lh5.ls(outfile, "/evt/")) == 11
    nda = {
        e: store.read(f"/evt/{e}", outfile)[0].view_as("np")
        for e in ["energy", "energy_aux", "energy_sum", "multiplicity"]
    }
    assert (
        nda["energy"][nda["multiplicity"] == 1]
        == nda["energy_aux"][nda["multiplicity"] == 1]
    ).all()
    assert (
        nda["energy"][nda["multiplicity"] == 1]
        == nda["energy_sum"][nda["multiplicity"] == 1]
    ).all()
    assert (
        nda["energy_aux"][nda["multiplicity"] == 1]
        == nda["energy_sum"][nda["multiplicity"] == 1]
    ).all()

    eid = store.read("/evt/energy_id", outfile)[0].view_as("np")
    eidx = store.read("/evt/energy_idx", outfile)[0].view_as("np")
    eidx = eidx[eidx != 999999999999]

    ids = store.read("hardware_tcm_1/array_id", lgnd_test_data.get_path(tcm_path))[
        0
    ].view_as("np")
    ids = ids[eidx]
    assert ak.all(ids == eid[eid != 0])


def test_lar_module(lgnd_test_data, tmptestdir):
    outfile = f"{tmptestdir}/l200-p03-r001-phy-20230322T160139Z-tier_evt.lh5"
    tcm_path = "lh5/prod-ref-l200/generated/tier/tcm/phy/p03/r001/l200-p03-r001-phy-20230322T160139Z-tier_tcm.lh5"
    if os.path.exists(outfile):
        os.remove(outfile)
    build_evt(
        f_tcm=lgnd_test_data.get_path(tcm_path),
        f_dsp=lgnd_test_data.get_path(tcm_path.replace("tcm", "dsp")),
        f_hit=lgnd_test_data.get_path(tcm_path.replace("tcm", "hit")),
        f_evt=outfile,
        evt_config=f"{config_dir}/module-test-evt-config.json",
        wo_mode="o",
        evt_group="evt",
        hit_group="hit",
        dsp_group="dsp",
        tcm_group="hardware_tcm_1",
    )

    assert os.path.exists(outfile)
    assert len(lh5.ls(outfile, "/evt/")) == 10
    nda = {
        e: store.read(f"/evt/{e}", outfile)[0].view_as("np")
        for e in ["lar_multiplicity", "lar_multiplicity_dplms", "t0", "lar_time_shift"]
    }
    assert np.max(nda["lar_multiplicity"]) <= 3
    assert np.max(nda["lar_multiplicity_dplms"]) <= 3
    assert ((nda["lar_time_shift"] + nda["t0"]) >= 0).all()


def test_lar_t0_vov_module(lgnd_test_data, tmptestdir):
    outfile = f"{tmptestdir}/l200-p03-r001-phy-20230322T160139Z-tier_evt.lh5"
    tcm_path = "lh5/prod-ref-l200/generated/tier/tcm/phy/p03/r001/l200-p03-r001-phy-20230322T160139Z-tier_tcm.lh5"
    if os.path.exists(outfile):
        os.remove(outfile)
    build_evt(
        f_tcm=lgnd_test_data.get_path(tcm_path),
        f_dsp=lgnd_test_data.get_path(tcm_path.replace("tcm", "dsp")),
        f_hit=lgnd_test_data.get_path(tcm_path.replace("tcm", "hit")),
        f_evt=outfile,
        evt_config=f"{config_dir}/module-test-t0-vov-evt-config.json",
        wo_mode="o",
        evt_group="evt",
        hit_group="hit",
        dsp_group="dsp",
        tcm_group="hardware_tcm_1",
    )

    assert os.path.exists(outfile)
    assert len(lh5.ls(outfile, "/evt/")) == 12
    nda = {
        e: store.read(f"/evt/{e}", outfile)[0].view_as("np")
        for e in ["lar_multiplicity", "lar_multiplicity_dplms", "lar_time_shift"]
    }
    assert np.max(nda["lar_multiplicity"]) <= 3
    assert np.max(nda["lar_multiplicity_dplms"]) <= 3

    ch_idx = store.read("/evt/lar_tcm_index", outfile)[0].view_as("ak")
    pls_idx = store.read("/evt/lar_pulse_index", outfile)[0].view_as("ak")
    assert ak.count(ch_idx) == ak.count(pls_idx)
    assert ak.all(ak.count(ch_idx, axis=-1) == ak.count(pls_idx, axis=-1))


def test_vov(lgnd_test_data, tmptestdir):
    outfile = f"{tmptestdir}/l200-p03-r001-phy-20230322T160139Z-tier_evt.lh5"
    tcm_path = "lh5/prod-ref-l200/generated/tier/tcm/phy/p03/r001/l200-p03-r001-phy-20230322T160139Z-tier_tcm.lh5"
    if os.path.exists(outfile):
        os.remove(outfile)
    build_evt(
        f_tcm=lgnd_test_data.get_path(tcm_path),
        f_dsp=lgnd_test_data.get_path(tcm_path.replace("tcm", "dsp")),
        f_hit=lgnd_test_data.get_path(tcm_path.replace("tcm", "hit")),
        f_evt=outfile,
        evt_config=f"{config_dir}/vov-test-evt-config.json",
        wo_mode="o",
        evt_group="evt",
        hit_group="hit",
        dsp_group="dsp",
        tcm_group="hardware_tcm_1",
    )

    assert os.path.exists(outfile)
    assert len(lh5.ls(outfile, "/evt/")) == 12
    vov_ene, _ = store.read("/evt/energy", outfile)
    vov_aoe, _ = store.read("/evt/aoe", outfile)
    arr_ac, _ = store.read("/evt/multiplicity", outfile)
    vov_aoeene, _ = store.read("/evt/energy_times_aoe", outfile)
    vov_eneac, _ = store.read("/evt/energy_times_multiplicity", outfile)
    arr_ac2, _ = store.read("/evt/multiplicity_squared", outfile)
    assert isinstance(vov_ene, VectorOfVectors)
    assert isinstance(vov_aoe, VectorOfVectors)
    assert isinstance(arr_ac, Array)
    assert isinstance(vov_aoeene, VectorOfVectors)
    assert isinstance(vov_eneac, VectorOfVectors)
    assert isinstance(arr_ac2, Array)
    assert (np.diff(vov_ene.cumulative_length.nda, prepend=[0]) == arr_ac.nda).all()

    vov_eid = store.read("/evt/energy_id", outfile)[0].view_as("ak")
    vov_eidx = store.read("/evt/energy_idx", outfile)[0].view_as("ak")
    vov_aoe_idx = store.read("/evt/aoe_idx", outfile)[0].view_as("ak")

    ids = store.read("hardware_tcm_1/array_id", lgnd_test_data.get_path(tcm_path))[
        0
    ].view_as("ak")
    ids = ak.unflatten(ids[ak.flatten(vov_eidx)], ak.count(vov_eidx, axis=-1))
    assert ak.all(ids == vov_eid)

    arr_ene = store.read("/evt/energy_sum", outfile)[0].view_as("ak")
    assert ak.all(arr_ene == ak.nansum(vov_ene.view_as("ak"), axis=-1))
    assert ak.all(vov_aoe.view_as("ak") == vov_aoe_idx)


def test_graceful_crashing(lgnd_test_data, tmptestdir):
    outfile = f"{tmptestdir}/l200-p03-r001-phy-20230322T160139Z-tier_evt.lh5"
    tcm_path = "lh5/prod-ref-l200/generated/tier/tcm/phy/p03/r001/l200-p03-r001-phy-20230322T160139Z-tier_tcm.lh5"
    if os.path.exists(outfile):
        os.remove(outfile)
    f_tcm = lgnd_test_data.get_path(tcm_path)
    f_dsp = lgnd_test_data.get_path(tcm_path.replace("tcm", "dsp"))
    f_hit = lgnd_test_data.get_path(tcm_path.replace("tcm", "hit"))
    f_config = f"{config_dir}/basic-evt-config.json"

    with pytest.raises(KeyError):
        build_evt(f_dsp, f_tcm, f_hit, outfile, f_config)

    with pytest.raises(KeyError):
        build_evt(f_tcm, f_hit, f_dsp, outfile, f_config)

    with pytest.raises(TypeError):
        build_evt(f_tcm, f_dsp, f_hit, outfile, None)

    conf = {"operations": {}}
    with pytest.raises(ValueError):
        build_evt(f_tcm, f_dsp, f_hit, outfile, conf)

    conf = {"channels": {"geds_on": ["ch1084803", "ch1084804", "ch1121600"]}}
    with pytest.raises(ValueError):
        build_evt(f_tcm, f_dsp, f_hit, outfile, conf)

    conf = {
        "channels": {"geds_on": ["ch1084803", "ch1084804", "ch1121600"]},
        "outputs": ["foo"],
        "operations": {
            "foo": {
                "channels": "geds_on",
                "aggregation_mode": "banana",
                "expression": "hit.cuspEmax_ctc_cal > a",
                "parameters": {"a": 25},
                "initial": 0,
            }
        },
    }
    with pytest.raises(ValueError):
        build_evt(f_tcm, f_dsp, f_hit, outfile, conf)


def test_query(lgnd_test_data, tmptestdir):
    outfile = f"{tmptestdir}/l200-p03-r001-phy-20230322T160139Z-tier_evt.lh5"
    tcm_path = "lh5/prod-ref-l200/generated/tier/tcm/phy/p03/r001/l200-p03-r001-phy-20230322T160139Z-tier_tcm.lh5"
    if os.path.exists(outfile):
        os.remove(outfile)
    build_evt(
        f_tcm=lgnd_test_data.get_path(tcm_path),
        f_dsp=lgnd_test_data.get_path(tcm_path.replace("tcm", "dsp")),
        f_hit=lgnd_test_data.get_path(tcm_path.replace("tcm", "hit")),
        f_evt=outfile,
        evt_config=f"{config_dir}/query-test-evt-config.json",
        wo_mode="o",
        evt_group="evt",
        hit_group="hit",
        dsp_group="dsp",
        tcm_group="hardware_tcm_1",
    )
    assert len(lh5.ls(outfile, "/evt/")) == 12


def test_vector_sort(lgnd_test_data, tmptestdir):
    outfile = f"{tmptestdir}/l200-p03-r001-phy-20230322T160139Z-tier_evt.lh5"
    tcm_path = "lh5/prod-ref-l200/generated/tier/tcm/phy/p03/r001/l200-p03-r001-phy-20230322T160139Z-tier_tcm.lh5"
    if os.path.exists(outfile):
        os.remove(outfile)
    f_tcm = lgnd_test_data.get_path(tcm_path)
    f_dsp = lgnd_test_data.get_path(tcm_path.replace("tcm", "dsp"))
    f_hit = lgnd_test_data.get_path(tcm_path.replace("tcm", "hit"))

    conf = {
        "channels": {"geds_on": ["ch1084803", "ch1084804", "ch1121600"]},
        "outputs": ["acend_id", "t0_acend", "decend_id", "t0_decend"],
        "operations": {
            "acend_id": {
                "channels": "geds_on",
                "aggregation_mode": "gather",
                "query": "hit.cuspEmax_ctc_cal>25",
                "expression": "tcm.array_id",
                "sort": "ascend_by:dsp.tp_0_est",
            },
            "t0_acend": {
                "aggregation_mode": "keep_at_ch:evt.acend_id",
                "expression": "dsp.tp_0_est",
            },
            "decend_id": {
                "channels": "geds_on",
                "aggregation_mode": "gather",
                "query": "hit.cuspEmax_ctc_cal>25",
                "expression": "tcm.array_id",
                "sort": "descend_by:dsp.tp_0_est",
            },
            "t0_decend": {
                "aggregation_mode": "keep_at_ch:evt.acend_id",
                "expression": "dsp.tp_0_est",
            },
        },
    }
    build_evt(f_tcm, f_dsp, f_hit, outfile, conf)

    assert os.path.exists(outfile)
    assert len(lh5.ls(outfile, "/evt/")) == 4
    vov_t0, _ = store.read("/evt/t0_acend", outfile)
    nda_t0 = vov_t0.to_aoesa().view_as("np")
    assert ((np.diff(nda_t0) >= 0) | (np.isnan(np.diff(nda_t0)))).all()
    vov_t0, _ = store.read("/evt/t0_decend", outfile)
    nda_t0 = vov_t0.to_aoesa().view_as("np")
    assert ((np.diff(nda_t0) <= 0) | (np.isnan(np.diff(nda_t0)))).all()


def test_tcm_id_table_pattern(lgnd_test_data, tmptestdir):
    outfile = f"{tmptestdir}/l200-p03-r001-phy-20230322T160139Z-tier_evt.lh5"
    tcm_path = "lh5/prod-ref-l200/generated/tier/tcm/phy/p03/r001/l200-p03-r001-phy-20230322T160139Z-tier_tcm.lh5"
    if os.path.exists(outfile):
        os.remove(outfile)
    f_tcm = lgnd_test_data.get_path(tcm_path)
    f_dsp = lgnd_test_data.get_path(tcm_path.replace("tcm", "dsp"))
    f_hit = lgnd_test_data.get_path(tcm_path.replace("tcm", "hit"))
    f_config = f"{config_dir}/basic-evt-config.json"

    with pytest.raises(ValueError):
        build_evt(f_tcm, f_dsp, f_hit, outfile, f_config, tcm_id_table_pattern="ch{{}}")
    with pytest.raises(ValueError):
        build_evt(f_tcm, f_dsp, f_hit, outfile, f_config, tcm_id_table_pattern="ch{}{}")
    with pytest.raises(NotImplementedError):
        build_evt(
            f_tcm, f_dsp, f_hit, outfile, f_config, tcm_id_table_pattern="ch{tcm_id}"
        )
    with pytest.raises(ValueError):
        build_evt(
            f_tcm, f_dsp, f_hit, outfile, f_config, tcm_id_table_pattern="apple{}banana"
        )



"""
This module implements routines to build the `skm` tier, consisting of skimmed
data from lower tiers.
"""

from __future__ import annotations

import json
import logging
import os

import awkward as ak
import numpy as np
from lgdo import Array, Table, lh5
from lgdo.lh5 import LH5Store

from pygama.evt import utils

log = logging.getLogger(__name__)


def build_skm(
    f_evt: str,
    f_hit: str,
    f_dsp: str,
    f_tcm: str,
    f_skm: str,
    skm_conf: dict | str,
    wo_mode="w",
    skm_group: str = "skm",
    evt_group: str = "evt",
    tcm_group: str = "hardware_tcm_1",
    dsp_group: str = "dsp",
    hit_group: str = "hit",
    tcm_id_table_pattern: str = "ch{}",
) -> None:
    """Builds a skimmed file from a (set) of `evt/hit/dsp` tier file(s).

    Parameters
    ----------
    f_evt
        path of `evt` file.
    f_hit
        path of `hit` file.
    f_dsp
        path of `dsp` file.
    f_tcm
        path of `tcm` file.
    f_skm
        name of the `skm` output file.
    skm_conf
        name of configuration file or dictionary defining `skm` fields.

        - ``multiplicity`` defines up to which row length
          :class:`.VectorOfVector` fields should be kept.
        - ``postfixes`` list of postfixes must be list of
          ``len(multiplicity)``. If not given, numbers from 0 to
          ``multiplicity -1`` are used
        - ``operations`` are forwarded from lower tiers and clipped/padded
          according to ``missing_value`` if needed. If the forwarded field
          is not an evt tier, ``tcm_idx`` must be passed that specifies the
          value to pick across channels.

        For example:

        .. code-block:: json

           {
             "multiplicity": 2,
             "postfixes":["", "aux"],
             "operations": {
               "timestamp":{
                 "forward_field": "evt.timestamp"
               },
               "multiplicity":{
                 "forward_field": "evt.multiplicity"
               },
               "energy":{
                 "forward_field": "hit.cuspEmax_ctc_cal",
                 "missing_value": "np.nan",
                 "tcm_idx": "evt.energy_idx"
               },
               "energy_id":{
                 "forward_field": "tcm.array_id",
                 "missing_value": 0,
                 "tcm_idx": "evt.energy_idx"
               }
             }
           }

    wo_mode
        writing mode.

        - ``write_safe`` or ``w``: only proceed with writing if the file does
          not already exists.
        - ``append`` or ``a``: append  to file.
        - ``overwrite`` or ``o``: replaces existing file.

    skm_group
        `skm` LH5 root group name.
    evt_group
        `evt` LH5 root group name.
    hit_group
        `hit` LH5 root group name.
    dsp_group
        `dsp` LH5 root group name.
    tcm_group
        `tcm` LH5 root group name.
    tcm_id_table_pattern
        pattern to format `tcm` id values to table name in higher tiers. Must have one
        placeholder which is the `tcm` id.
    """
    f_dict = {evt_group: f_evt, hit_group: f_hit, dsp_group: f_dsp, tcm_group: f_tcm}
    log = logging.getLogger(__name__)
    log.debug(f"I am skimming {len(f_evt) if isinstance(f_evt,list) else 1} files")

    tbl_cfg = skm_conf
    if not isinstance(tbl_cfg, (str, dict)):
        raise TypeError()
    if isinstance(tbl_cfg, str):
        with open(tbl_cfg) as f:
            tbl_cfg = json.load(f)

    # Check if multiplicity is given
    if "multiplicity" not in tbl_cfg.keys():
        raise ValueError("multiplicity field missing")

    multi = int(tbl_cfg["multiplicity"])
    store = LH5Store()
    # df = pd.DataFrame()
    table = Table()
    if "operations" in tbl_cfg.keys():
        for op in tbl_cfg["operations"].keys():
            miss_val = np.nan
            if "missing_value" in tbl_cfg["operations"][op].keys():
                miss_val = tbl_cfg["operations"][op]["missing_value"]
                if isinstance(miss_val, str) and (
                    miss_val in ["np.nan", "np.inf", "-np.inf"]
                ):
                    miss_val = eval(miss_val)

            fw_fld = tbl_cfg["operations"][op]["forward_field"]

            # load object if from evt tier
            if evt_group in fw_fld.replace(".", "/"):
                obj = store.read(
                    f"/{fw_fld.replace('.','/')}", f_dict[fw_fld.split(".", 1)[0]]
                )[0].view_as("ak")

            # else collect data from lower tier via tcm_idx
            else:
                if "tcm_idx" not in tbl_cfg["operations"][op].keys():
                    raise ValueError(
                        f"{op} is an sub evt level operation. tcm_idx field must be specified"
                    )
                tcm_idx_fld = tbl_cfg["operations"][op]["tcm_idx"]
                tcm_idx = store.read(
                    f"/{tcm_idx_fld.replace('.','/')}",
                    f_dict[tcm_idx_fld.split(".")[0]],
                )[0].view_as("ak")[:, :multi]

                obj = ak.Array([[] for x in range(len(tcm_idx))])

                # load TCM data to define an event
                ids = store.read(f"/{tcm_group}/array_id", f_tcm)[0].view_as("ak")
                ids = ak.unflatten(ids[ak.flatten(tcm_idx)], ak.count(tcm_idx, axis=-1))

                idx = store.read(f"/{tcm_group}/array_idx", f_tcm)[0].view_as("ak")
                idx = ak.unflatten(idx[ak.flatten(tcm_idx)], ak.count(tcm_idx, axis=-1))

                if "tcm.array_id" == tbl_cfg["operations"][op]["forward_field"]:
                    obj = ids
                elif "tcm.array_idx" == tbl_cfg["operations"][op]["forward_field"]:
                    obj = idx

                else:
                    chns = np.unique(
                        ak.to_numpy(ak.flatten(ids), allow_missing=False)
                    ).astype(int)

                    # Get the data
                    for ch in chns:
                        ch_idx = idx[ids == ch]
                        ct_idx = ak.count(ch_idx, axis=-1)
                        fl_idx = ak.to_numpy(ak.flatten(ch_idx), allow_missing=False)

                        if (
                            f"{utils.get_table_name_by_pattern(tcm_id_table_pattern,ch)}/{fw_fld.replace('.','/')}"
                            not in lh5.ls(
                                f_dict[[key for key in f_dict if key in fw_fld][0]],
                                f"ch{ch}/{fw_fld.rsplit('.',1)[0]}/",
                            )
                        ):
                            och = Array(nda=np.full(len(fl_idx), miss_val))
                        else:
                            och, _ = store.read(
                                f"{utils.get_table_name_by_pattern(tcm_id_table_pattern,ch)}/{fw_fld.replace('.','/')}",
                                f_dict[[key for key in f_dict if key in fw_fld][0]],
                                idx=fl_idx,
                            )
                        if not isinstance(och, Array):
                            raise ValueError(
                                f"{type(och)} not supported. Forward only Array fields"
                            )
                        och = och.view_as("ak")
                        och = ak.unflatten(och, ct_idx)
                        obj = ak.concatenate((obj, och), axis=-1)

            # Pad, clip and numpyfy
            if obj.ndim > 1:
                obj = ak.pad_none(obj, multi, clip=True)
            obj = ak.to_numpy(ak.fill_none(obj, miss_val))

            if obj.ndim > 1:
                if "postfixes" in tbl_cfg.keys():
                    nms = [f"{op}{x}" for x in tbl_cfg["postfixes"]]
                else:
                    nms = [f"{op}_{x}" for x in range(multi)]

                for i in range(len(nms)):
                    # add attribute if present
                    ob = Array(nda=obj[:, i])
                    if "lgdo_attrs" in tbl_cfg["operations"][op].keys():
                        ob.attrs |= tbl_cfg["operations"][op]["lgdo_attrs"]
                    table.add_field(nms[i], ob, True)
            else:
                obj = Array(nda=obj)
                if "lgdo_attrs" in tbl_cfg["operations"][op].keys():
                    obj.attrs |= tbl_cfg["operations"][op]["lgdo_attrs"]
                table.add_field(op, obj, True)

    # last thing missing is writing it out
    if wo_mode not in ["w", "write_safe", "o", "overwrite", "a", "append"]:
        raise ValueError(f"wo_mode {wo_mode} not valid.")
    log.debug("saving skm file")
    if (wo_mode in ["w", "write_safe"]) and os.path.exists(f_skm):
        raise FileExistsError(f"Write_safe mode: {f_skm} exists.")

    wo = wo_mode if wo_mode not in ["o", "overwrite"] else "of"
    store.write(obj=table, name=f"/{skm_group}/", lh5_file=f_skm, wo_mode=wo)
